#!/usr/bin/python

import globals

import pydb
import sys, re, time

import MPNG

import AlbDict
import Factors
import Number
import NumDenMul
import Result

import TripleStart
import StartTriple

import Saver
import FactorCache
import SigmaFactorCache
import ResultCache

import TimeStamp

RSLTS_txt="RSLTS.txt"
RSLTS_tmp="RSLTS.tmp"

verbose = 0

class ALB:

    def debug( self ):
        # Force a call to the debugger in running code here
#       pydb.debugger( )
        # ...
        pass

    def __init__( self, filename = None ):
        self.first_cycle = True
        self._mp_cnt = 0
        self._saver = Saver.Saver( self )
        self._starter = TripleStart.TripleStart( )

        if filename != None:
            self.load( filename )

    def doSave( self, force = False ):

        self._saver.doSave( force )

    def save( self, filename = RSLTS_tmp ):

        if globals.save == False:
            return

        if filename == None:
            file = sys.stdout
        else:
            t = TimeStamp.TimeStamp( )
            print "%s " % t,
            file = open( filename, "w" )
            print "Writing %s" % filename

        file.write( "%s\n" % self._starter )
        file.write( "%s\n" % globals.ResultCache.counts( ) )

        fmt = "\n*** %s ***\n%s\n**************\n"

        file.write( fmt % ( "factors", globals.FactorCache ) )
        file.write( fmt % ( "sgmftr",  globals.SigmaFactorCache ) )
        file.write( fmt % ( "rslts",   globals.ResultCache ) )

        if file != sys.stdout:
            file.close( )
            print "... done"
            sys.stdout.flush()

        return

    def load( self, filename = RSLTS_txt ):

        if globals.quiet == 0:
            beg_read = time.time( )
            print "Reading %s" % filename

        file = open( filename, "r" )

        sig = "tri"

        # process prefix
        while True:
            line = file.readline( )

            if line == None or line[0] == "*":
                break
            if line == "":
                continue

            parts = re.split( " +", line )
            if parts[ 0 ] == sig:
                self._starter.load( line )

        print "%s" % self._starter

        # process factors, sgmftr, and rslts
        while True:
            if line == None:
                line = file.readline( )
            if line == '':
                break
            if line == "" or line[0] != "*":
                line = None
                continue
            line = line.strip( )
            if verbose:
                print line
            parts = re.split( " +", line )
            line = None
            if len( parts ) != 3:
                continue

            beg = time.time( )

            if parts[ 1 ] == "factors":
                self.loading( parts[ 1 ] )
                globals.FactorCache.load( file )
                if verbose > 1:
                    print "%s" % fc

            if parts[ 1 ] == "sgmftr":
                self.loading( parts[ 1 ] )
                globals.SigmaFactorCache.load( file )
                if verbose > 1:
                    print "%s" % sfc

            if parts[ 1 ] == "rslts":
                self.loading( parts[ 1 ] )
                globals.ResultCache.load( file )
                if verbose > 1:
                    print "%s" % rc

            fin = time.time( )
            if verbose:
                print "time: %4.2f" % ( fin - beg )
            sys.stdout.flush()

        file.close( )

        if globals.quiet == 0:
            fin_read = time.time( )
            print "  %4.2fs" % ( fin_read - beg_read )
            sys.stdout.flush()
            print

        return

    def loading( self, tag ):
        print "%s loading %s" % ( TimeStamp.TimeStamp( ), tag )

    def cycle( self, max, seed ):
        """
        Answer the instance variable factorList.
        """

        ndm = NumDenMul.new( seed )

        self._seed = seed
        self._ndm  = ndm

        if globals.printState:
            print "ALB.cycle:  %s" % seed

        beg = time.time( )

        mpng = MPNG.MPNG( self )
        ( beg_cnt, net_cnt, ndm_dup, ndm_con ) = mpng.search( max, seed, ndm )

        fin = time.time( )

        if globals.qvs_flag > 1 or globals.printCycle or verbose - globals.quiet > 0:
            print "%5d %s %s %s" % ( beg_cnt, ndm.mul( ), ndm.num( ), ndm.den( ) )
        elif True:
            pass
        elif beg_cnt > 1000 or fin - beg > 10 or \
             ndm_dup > 100  or ndm_con > 100:
            print "bg:%5d, tm:%7.2f," % ( beg_cnt, fin - beg ),
            print "nd:%4d, nc:%4d," % ( ndm_dup, ndm_con ),
            print "m: %s" % ndm.mul( )

        if globals.qvs_flag and beg_cnt > 1000:
            print "cy: %5d,  mp: %2d,  st: %s" % ( beg_cnt, fnd_cnt, ndm.num( ) )

        return beg_cnt

    def adjMinMax_xxx_0928_xxxdef( self, seed ):
        """
        Multiply by all large primes to get ballpark exponents for 2, 3, 5,and 7
        """

        ndm = NumDenMul.new( seed )

        # reduce large ndm.num primes (to less than 10)
        factors = ndm.reduceLargeFactors( 10 )

        # search results for 'two' and determine
        # starter exponent ranges

        self._starter.adjMinMax( seed, factors )

        return

    def restart( self ):
        if self.first_cycle:
            self.first_cycle = False
            Factors.restart( )

    def saveMultiperfect( self, ndm ):
        """
        Save multiperfects.
        """

        mag = ndm.num( ).magnitude( )
        mul = ndm.mul( )

        result = Result.Result( mag, mul )
        result = globals.ResultCache.add( result, self._seed )

        state = result.state( )
        if state == "dup":
            return

        if globals.printDate:
            print "%s" % TimeStamp.TimeStamp( ),
        if globals.printDate or globals.printResult:
            print "%s 1 %s" % ( mag, mul )

        self._mp_cnt += 1
        print "#### %3d" % self._mp_cnt,

        if globals.quiet > 1:
            print "%s %s -> %s" % ( state, self._seed, result )
        else:
            if state != "new":
                tag = "####"
            else:
                tag = "####" * 5
            print tag
            print "#### %s %s -> %s" % ( result.state( ), self._seed, result )
            print tag

        result.state( "dup" )

        sys.stdout.flush()

        if state == "new":
            self.doSave( )

    # 1 - SearchBasedOnPower
    #  for results matching seed(2^N), 
    #      find min/max powers of 3,5,7,11,13,17,19
    #      extend max by 2
    #      with each prime/power combo as seed, search

    def SearchBasedOnPower( self, cnt, min, max ):

        while True:
            #self.search( cnt, min, max )
            ( min, seed ) = self.search( cnt, min, max )
            pydb.debugger( )
            if seed == None:
                break

        return

    # 2 - SearchBasedOnPowerAndDate
    #  for results matching seed(2^N),
    #      if 1 or more newer than 03/30,
    #	 find min/max powers of 3,5,7,11,13,17,19
    #	 extend max by 2
    #	 with each prime/power combo as seed, search

    def SearchBasedOnPowerAndDate( self, cnt, min, max ):

        pydb.debugger( )

        return ALB.SearchBasedOnPowerAndDate

    # 3 - SearchBasedOnLowCountAndFactorReduction
    #  for each N, find results matching seed(2^N)
    #      if count is 0 or 1,
    #      	 for p in 19,17,13,11,7,5,
    #	     reduce primes greater than p
    #	     find min/max powers of 3,5,7,11,13,17,19
    #         extend max by 2
    #	 with each prime/power combo as seed, search

    def SearchBasedOnLowCountAndFactorReduction( self, min, max ):

        powers = range( min, max )

        for power in powers:
#            print "SearchBasedOnLowCountAndFactorReduction( %s )" % power
            two = Number.Number( 2, power )
            print "########## %s ##########" % two

            seed = Factors.FromNumber( two )
            results_set = globals.ResultCache.matchFactors( seed )

            if len( results_set ) == 0:
                pydb.debugger( )
                self.SearchBasedOnFactorReduction( two, seed )
            else:
               self.SearchBasedOnLowCount( two, results_set )

            print ""

        return

    def SearchBasedOnLowCount( self, two, results_set ):

#       print "SearchBasedOnLowCount"

#       self._starter = TripleStart.TripleStart( "lc", two, seed )
        self._starter = TripleStart.TripleStart( )

        prime = two.prime( )
        power = two.power( )
        self._starter[ prime ].power( power )

        self._starter.mergeResults( power, results_set )
#       print self._starter.ranges_as_string( )
        self._starter.widen( 0, 3 )
#       print self._starter.ranges_as_string( )

        self.search( 0, 0, 0 )
#       self.search( 1, 0, 0 )

        return None

    def SearchBasedOnFactorReduction( self, two, seed ):

        print "SearchBasedOnFactorReduction"

        self._starter = TripleStart.TripleStart( "fr", two, seed )

#        triples = self._starter.findResultAndFactorReduce( seed )
#        triples.widen( 0, 2 )
#        self._starter.set_triples( two, triples )
#        self._starter.setRange( two.prime( ), two.power( ) )

        print "%s" % self._starter

        self.search( 1, 0, 0 )

        return None

    def search( self, cnt, min, max ):
        "Search for multiperfects."

        global printCycle

        self.restart( )
        sys.stdout.flush()

        total = 0
        cycle = min

        max += self._starter.count( )

        if globals.test_search == True:
#            seed = Factors.FromNumber( "2^2" )
#            self._starter.findResultMinMax( seed )

            pydb.debugger( )
            starter = Number.Number( 2, 2 )
            starter = Factors.FromNumber( starter )
            self._starter.findResultMinMax( starter )

#       print "cy: %d, mx: %d" % ( cycle, max )
#       sys.stdout.flush()

        self._starter.initialize( )

        while cycle < max or max == 0:

            beg = time.time( )

            try:
                ( test_count, seed ) = self._starter.next( )
#               print "n: %s" % seed
            except:
                pydb.debugger( )
                break

            if globals.qvs_flag > 1:
                print "s %3d %s" % ( cycle, seed )

            if globals.printCycle:
                print " "
                print "search %d %s" % ( cycle, seed )

            if seed == None:
                print "seed is None"
                break

            # new seed ...
            if cycle != 0 and len( seed ) == 1:
                break

            cycle += 1
            
            if globals.test_search == True:
                print "%4d %s" % ( cycle, seed )
                continue

            if globals.printCycle:
                print "%s" % ndm.asUString( )

            globals.quiet = 2
            count = self.cycle( cnt, seed )
            total += count

            fin = time.time( )
            secs = fin - beg

            if globals.qvs_flag or secs >= 100 or count > 10000:
                t = TimeStamp.TimeStamp( fin )
                print "%s %5d %5d %4ds  %s" % ( t, test_count, count, secs, seed )
                sys.stdout.flush()

            # save results (as needed)
            self.doSave( )

        print "count: %5d" % total
        sys.stdout.flush()

        return ( total, seed )

def Test( ):
    import Number
    mag = 1
    sgm = 1
    product = Number.Number( 1, 1 )
    for factor in ( "2^25", "3^4", "7^1", "11^2", "19^2", "127^1", "683^1", "2731^1", "8191^1" ):
        ( p, q ) = re.split( "\^", factor )
        n = Number.Number( p, q )
        s = n.sgmftr( ).magnitude( )
        sgm = sgm * s
        m = n.magnitude( )
        mag = mag * m
        print "%s   %s %s   %s %s" % ( n, m, mag, s, sgm )
    print "%s %s" % ( sgm / mag, sgm % mag )

def PrintVersion( ):
    print "Python %s" % sys.version
    print
    print TimeStamp.TimeStamp( )
    print

def do_letter_arg( arg, alb ):

    if arg == "T":
        Test( )
        
    if arg == "V":
        print sys.version
        
    if arg == "m":
       globals.testMode ^= True
        
    if arg == "d":
        testDebug += 1
        
    if arg == "q":
        globals.quiet += 1
        
    if arg == "v":
        verbose += 1
        
    if arg == "l":
        global filename
        alb.load( filename )
        
    if arg == "s":
        alb.doSave( True )
        
    if arg == "p":
        print globals.ResultCache
    
def do_search( arg, obj, method, cnt, min, max ):

    lc = arg.lower()

    # 1 - SearchBasedOnPower
    #  for results matching seed(2^N), 
    #      find min/max powers of 3,5,7,11,13,17,19
    #      extend max by 2
    #      with each prime/power combo as seed, search

    if lc == "s_p" or \
       lc == "s_power" or \
       arg == "SearchBasedOnPower":
        pydb.debugger( )
        return obj.SearchBasedOnPower( cnt, min, max )

    # 2 - SearchBasedOnPowerAndDate
    #  for results matching seed(2^N),
    #      if 1 or more newer than 03/30,
    #	 find min/max powers of 3,5,7,11,13,17,19
    #	 extend max by 2
    #	 with each prime/power combo as seed, search

    if lc == "s_pd" or \
       lc == "s_powerdate" or \
       arg == "SearchBasedOnPowerAndDate":
        pydb.debugger( )
        return obj.SearchBasedOnPowerAndDate( cnt, min, max )

    # 3 - SearchBasedOnLowCountAndFactorReduction
    #  for each N, find results matching seed(2^N)
    #      if count is 0 or 1,
    #      	 for p in 19,17,13,11,7,5,
    #	     reduce primes greater than p
    #	     find min/max powers of 3,5,7,11,13,17,19
    #         extend max by 2
    #	 with each prime/power combo as seed, search

    if lc == "s_lcfr" or \
       lc == "s_low_count_factor_reduction" or \
       arg == "SearchBasedOnLowCountAndFactorReduction":
        return obj.SearchBasedOnLowCountAndFactorReduction( min, max )

    pydb.debugger( )
    return method( cnt, min, max )

def main( ):

#    global quiet
#    global ndm_flag
#    global printCycle, printDate, printResult, printState
#    global testConflict, testDebug,globals.testMode

    beg = time.time( )

    globals.quiet = 0

    globals.verbose = 0
    globals.ndm_flag = 1	# NZ if _ndm_seen set used to trim search tree
    globals.qvs_flag = 0
    globals.test_search = False

    global filename
    filename = RSLTS_txt

    globals.printDate   = False # True
    globals.printResult = False # True
    globals.printState  = False # True
    globals.testMode    = False		# True if comparing old/new tree traversal
    cycleMax    = 0

    globals.printCycle  = False # True
    globals.testConflict= True  # False	# True if ndm.conflict() used to trim search tree
    globals.testDebug   = False # True

    alb = ALB( )
    mpng = MPNG.MPNG( alb )

    global t600, a600

    a600 = True
    t600 = True

    search = ALB.search

    if len( sys.argv ) == 1:

        PrintVersion( )

        globals.quiet += 2
        alb.load( )
        cnt = 0
        min = 1
        max = 1000000
        max = 0

        search( alb, cnt, min, max)

    else:

        factors = Factors.Factors( None )
        for arg in sys.argv[1:]:

            if arg == "new":  continue
            if arg == "old":  continue

            if len( arg ) == 1:
                do_letter_arg( arg, alb )
                continue

            if arg[ 0 ].isdigit( ):
                # "p" or "p^q"
                num = Number.FromString( arg )
                factors.add( num )
                continue

            arg12 = arg[ 0:2 ]

            if arg[ 0 ] == "f":
                # fq - q is 2's exponent
                q = arg[ 1: ]
                two = Number.Number( 2, q )
                verbose = 2
                triples = alb.createStartTriples( two )
                keys = triples.keys( )
                keys.sort( )
                for key in keys:
                    print triples[ key ],
                print
                continue
            
            if arg == "test_search":
                globals.test_search = True
                cnt = 3
                min = 1
                max = 1000
                search( alb, cnt, min, max)
                continue

            if arg12 == "cm":	# cm=NNN
                cycleMax = do_arg_with_incr( "cycleMax", cycleMax, arg )
                continue

            if arg12 == "pd":
                globals.printDate = do_arg_with_incr( "printDate", globals.printDate, arg )
                continue

            if arg12 == "pr":

                if arg == "pro":
                    # PrintResultsByOrder
                    globals.PrintResultsByOrder = True
                    continue

                if arg == "prv":
                    # PrintResultsByValue
                    globals.PrintResultsByOrder = False
                    continue

                printResult = do_arg_with_incr( "printResult", printResult, arg )
                continue

            if arg12 == "ps":
                printState = do_arg_with_incr( None, printState, arg )
                continue

            if arg12 == "pc":
                printCycle = do_arg_with_incr( "printCycle", printCycle, arg )
                continue

            if arg12 == "tc":
                testConflict ^= True
                continue

            if arg12 == "td":
                testDebug = do_arg_with_incr( "testDebug", testDebug, arg )
                continue

            if arg[0] == "r" or arg[0] == "R":
                if len( arg ) == 1:
                    arg = RSLTS_txt

                if len( arg ) != 2:
                    filename = arg
                    continue

                if arg[0:4] == "rslts" or  arg[0:4] == "RSLTS":
                    continue

                if arg == "r0":
                    printDate   = False
                    alb.load( )
                    cnt = 3
                    min = 1
                    max = 1000 # 10 200 500 1000
                    search( cnt, min, max)
                    return

                if arg == "rd1":
                    PrintVersion( )

                    arg = "r1"
#                   ndm_flag = 1
                    globals.quiet += 2
#                   printCycle   = False
#                   testConflict = True
#                   testDebug    = False

                if arg == "r1":
                    alb.load( )
                    cnt = 0
                    min = 1
                    max = 1000000
                    search( cnt, min, max)
                    continue

                if arg == "rw":
                    alb.load( filename )
                    alb.save( filename + ".new" )
                    return

            if arg[0] == "c":
                print "         %s" % ALB.instance
                print "ALB.fc:  %s" % ALB.instance._fc  #FactorCache( )
                print "ALB.sfc: %s" % ALB.instance._sfc #SigmaFactorCache( )
                
                print "ALB.FC:  %s" % FactorCache.instance( )
                print "ALB.SFC: %s" % SigmaFactorCache.instance( )
                print "ALB.RC:  %s" % ResultCache.instance( )
                continue

            if arg == "ndm":
                ndm_flag += 1
                continue

            if arg == "qvs":
                globals.qvs_flag += 1
                continue

            if arg12 == "s_":
                # s_abcd=min,max,cnt
                try:
                    tokens = re.split( "[=.,]*", arg )
                    arg = tokens[ 0 ]
                    min = int( tokens[ 1 ] )
                    max = int( tokens[ 2 ] )
                    cnt = int( tokens[ 3 ] )
                except:
                    min = 1
                    max = 20
                    cnt = 200

                do_search( arg, alb, search, cnt, min, max )

                alb.doSave( )

            if arg12 == "cy":
                printCycle = False
                globals.quiet += 2
                count = alb.cycle( cycleMax, factors)
                print "%d cycles" % count

            if arg[0] == "s":
                # s
                # s=min.max.cnt
                try:
                    tokens = re.split( "[s.=]*", arg )
                    min = int( tokens[ 1 ] )
                    max = int( tokens[ 2 ] )
                    cnt = int( tokens[ 3 ] )
                except:
                    min = 1
                    max = 20
                    cnt = 200

                alb.search( cnt, min, max )
                continue

            if arg == "x":
                # print results cache
                break

            if arg == "X":
                return

            if arg[ 0:5 ] == "a600:":
                a600 = arg[ 5: ] == 't'
                print "a600 = %s" % a600
                continue

            if arg[ 0:5 ] == "t600:":
                t600 = arg[ 5: ] == 't'
                print "t600 = %s" % t600
                continue

            print "unknown option - %s" % arg
            continue

        if len( factors.factors( ) ) != 0:
            printCycle = False
            globals.quiet += 2
#           testDebug = True
            count = alb.cycle( cycleMax, factors)
            print "%d cycles" % count

        fin = time.time( )
        tim = fin - beg
        if tim > 1:
            print "time: %4.2f" % tim

        print
        print globals.ResultCache.counts( )
        print

if __name__ == '__main__':

    main( )
