#!/usr/bin/python

import globals
import pydb
import re, time
from mpz import *

import Cache
import Factors

class FactorCache( Cache.Cache ):

    """
    Cache factor lists in a dictionary.
    Key is the value of the number.
    Value is the list of the numbers' factors (.
    """

    def __init__( self, saver = None ):

        Cache.Cache.__init__( self, saver )
        
    def type( self ):

        return "FactorCache"

    def __str__( self ):

        answer = ""
        keys = self.sortedkeys( )
        for key in keys:
            factors = self._dict[ key ]
            answer += "%s %s\n" % ( key, factors )

        if answer == "":
            answer = "<FactorCache.FactorCache instance>"

        return answer

    def addString( self, line ):

        tokens = re.split( "[()]+", line )
        value = mpz( tokens[ 0 ] )
        factors = Factors.FromString( tokens[ 1 ]  )
        self.addValue( value, factors, False )

    def addValue( self, value, factors, saver = True ):

        if self.verify( value, factors ):
            if not self._dict.has_key( value ):
                self.at_put( value, factors )
                if saver:
                    globals.Saver.newFactor( )

        return

    def verify( self, value, factors ):

        magnitude = factors.magnitude( )
        if value == magnitude:
            return True
        else:
            print "FactorCache error:"
            print "  %s" % value
            print "  %s" % factors
            pydb.debugger( )
            return False

if __name__ == '__main__':
    print "use Test_FactorCache.py for testing."
