#!/usr/bin/python

import globals
import pydb
import os, sys
import re, time, types
import math
from mpz import *
from random import randrange
import pyecm_wrapper

# 07/19/07 23:48 - LucasLehmer test timing
# time Factors.py mx,1000,2000 -- wi: 3.359s, wo: 12.258s
# time Factors.py mx,2000,3000 -- wi: 7.594s, wo: 37.177s

# Factors.py 1000000000000000001
# ( 101^1 9901^1 999999000001^1 )
#
#     07/23/07 21:21 - 3.64 s
#     12/31/07 12:26 - 0.207s

import AlbDict

import Number
import Wheel
import PrimeGenerator
import TimeStamp

verbose = 0
one_min   = 60
one_hour  = 3600

one_meg   = mpz( 1000000 )
one_gig   = mpz( 1000000000 )
ten_gig   = one_gig * mpz( 10 )
hun_gig   = one_gig * mpz( 100 )
ten_pow_15  = mpz( 10 ** 15 )
ten_pow_21  = mpz( 10 ** 21 )
state_file = "Factor_Status.txt"

def new( s ):

    f = Factors( s )
    return f

def restart( ):

    # read Factors_Status.txt and continue factoring (if needed)
    try:
        f = Factors( )
        ( val, pri ) = f.load_state( )
        val = mpz( val )
        pri = mpz( pri )
        fc = globals.FactorCache
        factors = fc.at( val )
        if factors == None:
            f.batchPrimeFactor( val, pri )
        else:
            pydb.debugger()
    except:
#       print "Error reading %s" % state_file
        pass

class Factors:

    """
    A dictionary of factors (Number objects).
    Key is the Number's prime.
    Value is the Number.
    """

    def __init__( self, n = None ):

        self._numbers = AlbDict.AlbDict( )
        self._str = None
        self._minimal = None

        if not n or n == 1:
            return

        beg = time.time( )
        self.factor( n )
        fin = time.time( )

        if fin - beg > 60:
            globals.Saver.doSave( )

    def type( self ):

        return "Factors"

    def __eq__( self, other ):

        if other == None:
            return False

        if len( self._numbers ) != len( other._numbers ):
            return False

        for key in self._numbers.keys( ):
            if not other._numbers.has_key( key ):
                return False
            if self._numbers[ key ] != other._numbers[ key ]:
                return False

        return True

    def __ne__( self, other ):

        return not ( self == other )

    def __cmp__( self, other ):

        sn = self._numbers.sortedvalues( )
        sl = len( sn )

        on = other._numbers.sortedvalues( )
        ol = len( on )

        for i in range( 0, min( sl, ol ) ):
            s = sn[ i ]
            o = on[ i ]
            if s != o:
                return s.__cmp__(o)

        return sl - ol

    def has_key( self, key ):

        return self._numbers.has_key( key )

    def __hash__( self ):

        v = 1
        for n in self._numbers.values( ):
            p = n.prime( )
            q = n.power( )
            q = min( q, 10 )
            v = int( ( v * p ** q ) % 1279 )

        return v

    def __getitem__( self, item ):

        if item == 0:
            pydb.debugger( )

        try:
            answer = self._numbers[ item ]
        except:
            pydb.debugger( )
            answer = self._numbers[ long( item ) ]

        return answer

    def __delitem__( self, item ):

        del self._numbers[ item ]

    def __len__( self ):

        return len( self._numbers )

    def __mul__( self, anObject ):

        # multiply self by 1 factor or by a factorlist

        result = self.copy( )

        if anObject.__class__ == Number.Number:
            # .. a factorlist
            factor = anObject
            fp = factor.prime( )
            fq = factor.power( )

            if result._numbers.has_key( fp ):
                sq = result._numbers[ fp ].power( )
                result._numbers[ fp ].power( sq + fq )
            else:
                result.add( factor )
        else:
            if anObject.__class__ == Factors:
                # ... a factor
                factors = anObject.factors( )
            else: # SigmaFactor
                factors = anObject.factors( ).factors( )

            for factor in factors:
                result.add( factor )

        return result

    def __div__( self, anObject ):

        # multiply self by 1 factor or by a factorlist

        result = self.copy( )

        if anObject.__class__ != Number.Number:
            # ... a factor
            for factor in anObject.factors( ):
                result = result / factor
        else:
            # .. a factorlist

            factor = anObject
            fp = factor.prime( )
            fq = factor.power( )

            if result._numbers.has_key( fp ):
                sq = result._numbers[ fp ].power( )
                if sq != fq:
                    result._numbers[ fp ].power( sq - fq )
                else:
                    del result._numbers[ fp ]
            else:
                print "*** Factors.div error ***"
                print "              num: %s" % result
                print "              den: %s" % anObject
                pydb.debugger( )
                return False

        return result

    def __str__( self ):

        return self.asString( )

    def asString( self ):

        if self._str == None:
            answer = ""
            for f in self.factors( ):
                answer += "%s " % f.asString( )
            if not isinstance( self._minimal, types.BooleanType ):
                tag = ""
            else:
                tag = "* "
            answer = "( %s%s)" % ( answer, tag )
            self._str = answer 

        return self._str

    def asUString( self ):

        answer = "( "
        for f in self.factors( ):
            answer += f.asUString( )
        answer += ")"

        return answer

    def factors( self ):
        # return a sorted list of p,q pairs

        answer = self._numbers.values( )
        if len( answer ) > 0:
            answer.sort( )

        return answer

    def at( self, index ):

        return self._numbers[ index ]

    def add( self, aFactor ):

        p = aFactor.prime( )

        if self._numbers.has_key( p ):
            q = aFactor.power( ) + self._numbers[ p ].power( )
            aFactor = Number.Number( p, q )

        self._numbers[ p ] = aFactor

    def addString( self, str ):

        ( p, q ) = re.split( "[\^]", str )
        p = mpz( p )
        q = mpz( q )
        aFactor = Number.Number( p, q )
        self.add( aFactor )

    def copy( self ):

        # answer a copy of self

        new = Factors( None )
        for number in self._numbers.values( ):
            new.add( number.copy( ) )

        return new

    def factor( self, anInteger ):

        # Return a dictionary with the p:q factors of anInteger

        v = mpz( anInteger )

        if v > one_meg:
            factors = globals.FactorCache.at( v )

            if factors != None:
                self._numbers = factors._numbers.copy( )
                return

            if v > ten_gig and not globals.quiet:
                print "factoring: %s" % v,
                sys.stdout.flush()

        if v < ten_pow_15:
            r = self.batchPrimeFactor( v )
        else:
#           r = self.pollard_rho_factor( v )
            r = self.pyecm( v )
            globals.Saver.doSave( True )

        self._numbers = r

        if anInteger > one_meg:
            if not globals.quiet:
                print
                print "%s" % self
                sys.stdout.flush()
            globals.FactorCache.add( anInteger, self )

        return

    def hasPrime( self, prime ):

        return self._numbers.has_key( prime )

    def includes( self, factors ):

#        pydb.debugger( )
        for factor in factors.values( ):
            prime = factor.prime( )
            if not self.has_key( prime ):
                return False
            if self[ prime ] != factor:
                return False

        return True

    def minimal( self, bool ):

        self._minimal = bool
        self._str = None

        return

    def nextPrimeFactor( self, val ):

        gen = PrimeGenerator.PrimeGenerator( 3 )

        pri = mpz( 1 )
        zer = mpz( 0 )
        dis = one_meg

        ans = { }    # dict
        
        tim = time.time( )

        self._trace_cnt = 0

        while True:
            pri = gen.NextPrime( )

            if pri * pri > val:
                break

            if val % pri == zer:
                exp = 0
                while val % pri == 0:
                    exp += 1
                    val = val / pri

                ans[ pri ] = Number.Number( pri, exp )

                if v > one_meg:
                    if self.test_cached_value( val, ans ):
                        break

            if pri > dis:
                ( dis, tim ) = self.trace_view( tim, val, pri, dis )

        if val > 1:
            ans[ val ] = Number.Number( val, 1 )

        return ans

    def batchPrimeFactor( self, val, pri = None ):

        ans = { }    # dict
        dis = one_meg
        gen = PrimeGenerator.PrimeBatchGenerator( 4 )
        zer = mpz( 0 )

        if not pri:
            pri = mpz( 1 )
        else:
            # use wheel (skip spokes)
            gen.wheel( )

        tim = time.time( )

        self._trace_cnt = 0
        self._next_trace = one_min

        while val > 1 and pri * pri <= val:

            ( val, pri ) = self.doPrimeBatch( val, ans, gen, pri, zer )

            if pri > dis:
                ( dis, tim ) = self.trace_view( tim, val, pri, dis )

        if val > 1:
            ans[ val ] = Number.Number( val, 1 )

        return ans

    def doPrimeBatch( self, v, r, g, p, z ):

        primes = g.primes( )

        for d in primes:
            p += d

            if v % p != z:
                continue
            q = 0
            while v % p == 0:
                q += 1
                v = v / p

            r[ p ] = Number.Number( p, q )
            if q != 0 and v > one_meg:
                if self.test_cached_value( v, r ):
                    v = 1
                    break

        return ( v, p )

    def test_cached_value( self, v, r ):

        factors = globals.FactorCache.at( v )
        if not factors:
            return False
        
        for factor in factors.values( ):
            p = factor.prime( )
            if r.has_key( p ):
                print "**********"
                print "rpt: %s" % p
                print "**********"
                sys.stdout.flush()
                pydb.debugger( )

            r[ p ] = factor.copy( )

        return True

    def first( self ):

        p = min( self._numbers.keys( ) )

        return self._numbers[ p ]

    def last( self ):

        try:
            p = max( self._numbers.keys( ) )
            l = self._numbers[ p ]
            return l
        except:
            return None

    def magnitude( self ):
        v = 1
        for n in self._numbers.values( ):
            m = n.magnitude( )
            v = v * m

        return v

    def matchesFactors( self, factors ):

        for factor in factors.values( ):
            p = factor.prime( )
            try:
                value = self._numbers[ p ]
                if value.power( ) != factor.power( ):
                    return False
            except:
                return False

        return True
        
    def detectPrime( self, p ):

        if self._numbers.has_key( p ):
            return True
        else:
            return False

    def selectLessThan( self, rdr ):

        result = Factors( None )
        for factor in self._numbers.values( ):
            if factor < rdr:
                result.add( factor )

        return result

    def remove( self, p ):

        del self._numbers[ p ]
        self._str = None

    def values( self ):

        return self._numbers.values( )

    def sortedvalues( self ):

        if isinstance( self._numbers, AlbDict.AlbDict ):
            return self._numbers.sortedvalues( )
        else:
            values = self._numbers.values( )
            values.sort( )
            return values

    def isPrime( self ):

        if len( self._numbers ) == 1 and self._numbers.values()[0].isPrime( ):
            return True
        else:
            return False

    def sgmftr( self ):

        answer = Factors( )
        values = self.sortedvalues( )
        for factor in values:
            sgmftr = factor.sgmftr( )
            answer *= sgmftr 

        return answer

    def magnitude( self ):

        answer = 1
        for factor in self.sortedvalues( ):
            try:
                magnitude = factor.magnitude( )
            except:
                pydb.debugger( )                
                magnitude = factor.magnitude( )
            answer *= magnitude

        return answer

    def trace_view( self, tim, val, pri, dis ):

        cur = time.time( )
        if cur > tim and globals.quiet <= 2:

            self.save_state( val, pri )

            pri = pri * 1.0
            sqrt = mpz( val ** 0.50 )

            tim = "%s" % TimeStamp.TimeStamp( )

            if dis == one_meg:
                print "%s  %d %8.3e  %d %8.3e" % ( tim, long( val ), val, long( sqrt ), sqrt )
                tim = " " * ( len ( tim ) - 5 )
                str = "%8.2f Mb" % ( pri / one_meg )
                lbl = "Mb"
                div = one_meg
            else:
                lbl = "Gb"
                div = one_gig
                str = "%8.2f Gb" % ( pri / one_gig )
            fmt = "  %s  %s %8.2f %7.2f"
            print fmt % ( tim, str, val / pri / pri, pri / sqrt * 100.0 )
#            fmt = "%s  %8.2f %s %8.2f %7.2f"
#            print fmt % ( tim, pri / div, lbl, val / pri / pri, pri / sqrt * 100.0 )
            sys.stdout.flush( )

            self._trace_cnt += 1
            if self._trace_cnt == 5:
                self._next_trace = one_hour

            tim = cur + self._next_trace

        if not globals.quiet:
            if dis == one_meg:
                print "factoring: %s" % val
            print ".",
            sys.stdout.flush()
            globals.Saver.doSave( )

        dis += one_meg

        return ( dis, tim )

    def load_state( self ):

        file = open( state_file, "r" )
        line = file.readline( )
        file.close( )

        ( val, pri ) = re.split( "[ ,]+", line )

        return ( val, pri )

    def save_state( self, val, pri ):

        if pri > one_gig:
            name = "%s.txt" % val
            file = open( name, "w" )
            file.write( "%s %s\n" % ( val, pri ) )
            file.close( )

    def power2mod( self, val, mer, exp ):
        rslt = ( val & mer ) + ( val >> exp )
        if rslt >= mer:
            rslt -= mer
        if rslt >= mer:
            rslt -= mer
            print "power2mod:  %d  %d  %d" % ( val, mer, exp )
        return rslt

    def isLucasLehmerPrime( self, q ):

        # http://primes.utm.edu/mersenne/index.html#test

        # Lucas-Lehmer Test: For p an odd prime, the Mersenne number 2^p-1 is
        # prime if and only if 2^p-1 divides S(p-1) where S(n+1) = S(n)^2-2,
        # and S(1) = 4. [Proof.]

        if not Factors( q ).isPrime( ):
            return False

        beg = time.time( )
        msg = None

        v = mpz( 2 ) ** q - 1
        s = mpz( 4 )

        for i in range( 2, q ):
            s = s ** 2 - 2
#           s = s % v
            s = self.power2mod( s, v, q )
        fin = time.time( )

        if s != 0:
            result = 0
            if verbose > 1:
                msg = "--"
        else:
            result = v
            if verbose > 0:
                msg = "mp"

        if msg != None:
            tim = TimeStamp.TimeStamp( "p" )
            print "  %s %8.2f %5d %s" % ( tim, fin - beg, q, msg )
            sys.stdout.flush()

        return result

    def pyecm( self, value ):

        fvalue = value * 1.0
        if fvalue > 1.0e20:
             print "pyecm:  %e  %s" % ( fvalue, value )
             sys.stdout.flush()
        beg = time.time( )

        ans = pyecm_wrapper.factors_as_dict_of_numbers( value )
        if fvalue > 1.0e20:
             print "pyecm:  %e  %s" % ( fvalue, value )
             sys.stdout.flush()

        fin = time.time( )
        tim = TimeStamp.TimeStamp( "p" )
        print "  %s %8.2f %s" % ( tim, fin - beg, ans )

        return ans

    def pollard_rho_factor( self, value ):
        return self.pollard_rho_factors_gmpy( value )

    def pollard_rho_factors_gmpy( self, value ):
        
        fvalue = value * 1.0
        if fvalue > 1.0e20:
             print "pollard_rho_factors_gmpy:  %e  %s" % ( fvalue, value )
             sys.stdout.flush()

#       pysymbolicext.verbose( );
        beg = time.time( )

        factors = pysymbolicext.factor( value )
        ans = { }    # dict
        for factor in factors:
            ( p, q ) = factor
            if ans.has_key( p ):
                q += ans[ p ].power( )
            ans[ p ] = Number.Number( p, q )

        if fvalue > 1.0e20:
             print "%s" % ans
             sys.stdout.flush()

        fin = time.time( )
        tim = TimeStamp.TimeStamp( "p" )
        print "  %s %8.2f %s" % ( tim, fin - beg, ans )

        return ans

    def print_factors( self, value, factors, tim = 0 ):

        # print factors dictionary

        print "%s (" % value,
        keys = factors.keys( )
        keys.sort( )
        for key in keys:
            print "%s" % factors[ key ],
        if tim < 0.1:
            print ")"
        else:
            print ")", " *** %5.2f secs" % tim

    def pollard_rho_factors_python( self, value ):

#        print "pollard_rho_factors_python %s" % value
        ans = { }
        starts = [ 2, 3, 4, 6 ]
        while ( value > 1 ):
            f = self.pollard_rho( starts, value )
            if f == 1:
                f = value
            q = 0
            print value, f, value / f
            pydb.debugger( )
            while ( value % f ) == 0:
                q += 1
                value = value / f
            ans[ f ] = Number.Number( f, q )

        return ans

    def pollard_rho( self, starts, value ):

        """
        factorPR(n) - Find a factor of n using the Pollard Rho method.
        Note: This method will occasionally fail.
        """
    #   http://www.math.umbc.edu/~campbell/Computers/Python/numbthy.py

        one = mpz( 1 )
        intrvl = one_meg
        for start in starts:
            slow = mpz( start )
            numsteps = 2 * math.floor( math.sqrt( math.sqrt( value ) ) )
            numsteps += 1000
            numsteps = mpz( numsteps  )
            fast = slow
            i = one
            d = intrvl
            while i < numsteps:
                if i == d:
                    d = self.rho_status_display( value, numsteps, start, i, intrvl )

                slow = ( slow * slow + one ) % value
                i = i + one
                fast = ( fast * fast + one ) % value
                fast = ( fast * fast + one ) % value
                g = self.gcd( max( fast, slow ) - min( fast, slow ), value )
                if ( g != one ):
                    if ( g == value ):
#                        print "g == value **** break"
                        break
                    else:
                        if g > one_meg or start != 2:
                            print TimeStamp.TimeStamp( ),
                            print "*** pollard_rho_factor:  s %d, g %d" % ( start, g )
                        return g
#        print "-> 1"
        return 1

    def egypt_gen( self, n, c=1 ):
        """
        Generate sequence x_i = (x_{i-1}^2 + c) mod n
        Where n is the target composite we want to factor.
        """
        x = 1
        while True:
            x = ( x ** 2 + c ) % n
            yield x

    def egypt_rho( self, n, maxt=500, maxc=10 ):
        """
        Pollard's Rho method for factoring n.
        Returns a list of factors (not necessarily prime) of n.
        Tests each polynomial x^2+c (c in range(1,maxc))
        by following the sequence gen(n,c) for maxt steps.
        If the sequence is cyclic modulo a factor of n
        with a smaller cycle length than its cycle modulo n,
        we can identify a factor of n as the gcd of n
        with the difference of two sequence values separated
        by the smaller cycle length in the sequence.
        """
#        if millrab(n):  # don't bother with probable primes
#            return [n]

        maxt = min( maxt, n )
        for c in range( 1, maxc ):
            results = self.egypt_rho_inner( n, c, maxt, maxc )
            if results and len( results ) != 1:
                return results

        return [ n ]    # failure to factor

    def egypt_rho_inner( self, n, c, maxt, maxc ):
        seqslow = self.egypt_gen( n, c ) 
        seqfast = self.egypt_gen( n, c )
        for trial in range( maxt ):
            xb = seqslow.next( )     # slow generator goes one step
            seqfast.next( )
            xk = seqfast.next( )     # while fast generator goes two
            diff = abs( xk-xb )
            if not diff:
                if globals.quiet <= 1:
                    print "n -> %d" % n
                return None # [ n ]
#                continue
            d = self.egypt_gcd( diff,n )         # have a factor?
            if n > d > 1:                
                if globals.quiet <= 1:
                    print "%s -> ( %s %s )" % ( n, d, n / d )
                return [ d, n // d ]
        return [ n ]

    # Find all factors and divisors
    # Fix rho (doesn't handle powers of two correctly) and speed up small primes

    def egypt_factors( self, n ):
        factors = self.egypt_factor_list( n )
        result = { }
        for f in factors:
            result[ f ] = result.get( f, 0 ) + 1
        for r in result.keys( ):
            result[ r ] = Number.Number( r, result[ r ] )
        return result

    def egypt_factor_list( self, n ):
        """
        Return a list of the factors of n.
        Uses trial division for small primes before switching to Pollard's Rho method.
        """
        f = [ ]
        for p in [ 2, 3, 5, 7, 11, 13, 17, 19 ]:
            q = 0
            while n % p == 0:
                n //= p
                q += 1
            if q != 0:
                f.append( p )
        if n == 1:
            return f
        if self.millrab( n ):
            return f + [n]
        maxt, maxc = 100, 5
        rho_results = [ n ]
        r = 1
        while len( rho_results ) == 1:
            beg = time.time( )
            rho_results = self.egypt_rho( n, maxt, maxc)
            fin = time.time( )
            if globals.quiet <= 1:
                print "n %s, r %d, t %5.2f" % ( n, r, fin - beg )
            maxt *= 2
            maxc *= 2 
            r = r * 2
        if globals.quiet <= 1:
            print rho_results
        for factor in rho_results:
            f += self.egypt_factor_list( factor )
        return f

    def millrab( self, n, max=30 ):
        """
        Miller-Rabin primality test as per the following source:
        http://www.wikipedia.org/wiki/Miller-Rabin_primality_test
        Returns probability p is prime: either p = 0 or ~1,
        """
        if not n % 2: return 0
        k = 0
        z = n - 1

        # compute m,k such that (2**k)*m = n-1
        while not z % 2:
          k += 1
          z //= 2
        m = z

        # try tests with max random integers between 2,n-1
        ok = 1
        trials = 0
        p = 1
        while trials < max and ok:
            a = randrange( 2, long( n ) - 1 )
            trials += 1
            test = pow( a, m, n )
            if ( not test == 1 ) and not ( test == n - 1 ):
                # if 1st test fails, fall through
                ok = 0
                for r in range( 1, k ):
                    test = pow( a, ( 2 ** r )* m, n )
                    if test == ( n - 1 ):
                        ok = 1 # 2nd test ok
                        break
            else: ok = 1  # 1st test ok
            if ok == 1:  p *= 0.25

        if ok:  return 1 - p
        else:   return 0


# http://en.wikipedia.org/wiki/Pollard_rho
# Richard Brent's variant

    def pollard_brent_factors( self, value ):
        ans = { }
        while ( value > 1 ):
            f = self.pollard_brent( value )
            if not f or f == 1:
                f = value
            q = 0
            while ( value % f ) == 0:
                q += 1
                value = value / f
            ans[ f ] = Number.Number( f, q )

        return ans

    def pollard_brent( self, value ):
        n = value
        m = 1

#####1. y = x0, r = 1, q = 1.
        y = mpz( 5 )
        r = 1
        q = 1

#####2. Do:
        g = 1
        while g == 1:
###########1. x = y
            x = y
###########2. For i = 1 To r:
###########      1. y = f(y)
            for i in range( 0, r ):
                y = ( y * y + 1 ) % n
###########3. k = 0
            k = 0
            g = 1
            while ( k < r and g == 1 ):
###########4. Do:
###########      1. ys = y
                ys = y
###########      2. For i = 1 To min(m, r - k):
                for i in range( 0, min( m, r - k ) ):
###########            1. y = f(y)
                    y = ( y * y + 1 ) % n
###########            2. q = (q * abs(x - y)) mod n
                    q = ( q * abs( x - y) ) % n
###########      3. g = GCD(q, n)
                g = self.gcd( q, n )
###########      4. k = k + m
                k = k + m
###########5. Until (k >= r or g > 1)
###########6. r = 2r
            r = 2 * r
            y = ( x * x + 1 ) % n ### DMR ###
#####3. Until g > 1

#####4. If g = n then
        if g == n:
            g = 1
            while g <= 1:
###########1. Do:
###########      1. ys = f(ys)
                ys = ( ys * ys + 1 )  % n
###########      2. g = GCD(abs(x - ys), n)
                g = self.gcd( abs( x - ys ), n )
###########2. Until g > 1

#####5. If g = n then return failure, else return g
        if g != n:
            return g
        else:
            return None

    def gcd( self, a, b ):
        return self.egypt_gcd( a, b )

    def rho_gcd( self, a, b ):

        """
        gcd( a,b ) returns the greatest common divisor of the integers a and b.
        """

        if a == 0:
            return b
        return abs( self.gcd( b % a, a ) )

    def egypt_gcd( self, a, b):
        """
        Euclid's algorithm for integer greatest common divisors.
        """

        while b:
            a, b = b, a % b
        return a

    def rho_status_display( self, val, numsteps, start, dis, intrvl ):

        tim = TimeStamp.TimeStamp( )
        tim = "%s" % tim

        if dis == intrvl:
            print "%s %d %8.3e  %d %8.3e" % ( tim, long( val ), val, long( numsteps ), numsteps )
            tim = " " * len( tim )

        if dis < one_gig:
            str = "%8.2f Mb" % ( dis * 1.0 / one_meg )
        else:
            str = "%8.2f Gb" % ( dis * 1.0 / one_gig )

        print "%s %s %s %9.4f%%" % ( tim, start, str, dis * 100.0 / numsteps )
        sys.stdout.flush( )

        max_interval = 100 * 1000 * 1000
        if intrvl <= max_interval:
            dis += intrvl
        else:
            dis += max_interval

        return dis

def FromNumber( aNumber ):

    p = mpz( aNumber.prime( ) )
    q = mpz( aNumber.power( ) )
    n = Number.Number( p, q )
    
    result = Factors( None )
    result.add( n )
    return result

def FromString( aString ):

    result = Factors( None )
    tokens = re.split( "[ ()]+", aString )

    for token in tokens:
        if len( token ) < 3:
            continue
        if token.find( "^" ) < 0:
            n = mpz( token )
            result = Factors( n )
        else:
            ( p, q ) = re.split( "[\^]", token )
            p = mpz( p )
            q = mpz( q )
            n = Number.Number( p, q )
            result.add( n )

    return result

def main( ):

    globals.quiet = 1
    globals.verbose = 0

    for arg in sys.argv[1:]:

        if arg[ 0 ].isdigit( ):
            f = FromString( arg )
            print "%s %s" % ( arg, f )

        if arg[ 0 ].isalpha( ):
            if arg == 'f':
                state_file = arg[ 1: ]
                restart( )
                continue

            if arg == 'l':
                globals.PROG.load( )
                continue

            if arg == 's':
                globals.PROG.save( )
                continue

# ./Factors.py l py=2607480390484525739   py=312145323578466047987495401
# ./Factors.py l gmpy=2607480390484525739 gmpy=312145323578466047987495401

            f = None
            factors = None
            tokens = re.split( "[:=]", arg )
            r = None
            b = time.time( )
            if len( tokens ) < 0:
                f = Factors( arg )
                factors = f.factors( )
            elif tokens[ 0 ] == 'q':
                globals.quiet += 1
            elif tokens[ 0 ] == 'sm':
                f.sgmftr_with_max( 12 )
            elif tokens[ 0 ] == 'gcd':
                f = Factors( None )
                print f.rho_gcd(   mpz( tokens[ 1 ] ), mpz( tokens[ 2 ] ) ),
                print f.egypt_gcd( mpz( tokens[ 1 ] ), mpz( tokens[ 2 ] ) )
            elif tokens[ 0 ] == 'er':
                f = Factors( None )
                v = mpz( tokens[ 1 ] )
                r = f.egypt_rho( v )
            elif tokens[ 0 ] == 'ef':
                if globals.quiet <= 1:
                    print tokens[1]
                f = Factors( None )
                v = mpz( tokens[ 1 ] )
                r = f.egypt_factors( v )

            elif tokens[ 0 ] == 'r':
                values = [ 12970571, 13209011, 17728931, 23845791,
                           12970571 * 13209011, 17728931 * 23845791,
                           12970571 * 13209011 * 17728931 * 23845791 ]
                v1, v2, v3, v4 = 12970573, 13209011, 17728931,  23845791
                v1, v2, v3, v4 = 131946889, 319020217, 412239473, 420778751
                values = [ # v1, v2, v3, v4,
                           # v1 * v2, v3 * v4,
                           v1 * v2 * v3 * v4,
                           v1 * v2 * v3 * v4 * 1000 + 1 ]
#                values = [ v1, v2, v3, v4 ]
                for v in values:
                    f = Factors( None )
                    b = time.time( )
                    r = f.egypt_factors( mpz( v ) )
                    f = time.time( )
                    print "n %s, r %s, t %5.2f" % ( v, r, f - b )
            else:
                f = Factors( None )
                ( m, v ) = tokens
                v = mpz( v )
                if m == 'py' or m == 'python':
                    r = f.pollard_rho_factors_python( v )
                if m == 'rho' or m == 'gmpy':
                    r = f.pollard_rho_factors_gmpy( v )
                if m == 'br' or m == 'pollard_brent':
                    r = f.pollard_brent_factors( v )

            if r:
                f.print_factors( v, r, time.time() - b )

if __name__ == '__main__':
    main( )
