#!/usr/bin/python

import globals

import pydb
import sys, re, time

import ALB
import AlbDict
import Factors
import Number
import NumDenMul
import Result

import TripleStart
import StartTriple

import Saver
import FactorCache
import SigmaFactorCache
import ResultCache

import TimeStamp

RSLTS_txt="RSLTS.txt"

verbose = 0

global isr, a600
isr = True # False # enable is_repeated_state
t600 = False
a600 = False

class MPNG_STAGE:

    def __init__( self, ndm, cur, rdr ):
        self._ndm = ndm.copy( )
        self._cur = cur
        self._rdr = rdr.copy( )
        return

    def type( self ):

        return "MPNG_STAGE"

    def __str__( self ):
        c = "%s" % self._cur
        n = "%s" % self._ndm.num( )
        d = "%s" % self._ndm.den( )
        m = "%s" % self._ndm.mul( )
        return "%s,  %s,  %s,  %s" % ( c, n, d, m )

    def value( self ):
        return ( self._ndm,  self._cur, self._rdr )

class MPNG:

    def debug( self ):
        # Force a call to the debugger in running code here
#       pydb.debugger( )
        # ...
        pass

    def __init__( self, alb, filename = None ):
        self._owner = alb
        self._mp_cnt = 0
        self._trace_cycle = 0
        self._n_cur = None
        self._n_rdr = None

        global isr
        if isr:
            self.state_seen = set( )
        self._reduce_max = 1000

        if filename != None:
            self.load( filename )

    def type( self ):

        return "MPNG"

    def search( self, max, seed, ndm = None ):

        global isr, t600

        self._o_fs = [ ]    # list
        self._n_fs = [ ]    # list
        self._level = 0
        self._indent = ""

        beg_cnt = 0
        net_cnt = 0
        ndm_dup = 0
        ndm_con = 0

        state = MPNG.a200

        if ndm == None:
            ndm = NumDenMul.new( seed )

        self._o_ndm = ndm.copy( )
        self._n_ndm = ndm.copy( )
        self._seed = seed

        while state != None and net_cnt < 100000:
            net_cnt += 1

            if self._trace_cycle != 0 and self._trace_cycle == net_cnt:
                self.debug( )

            if state == MPNG.a200:
                beg_cnt += 1
                
                if self.testConflict( ):
                    ndm_con += 1
                    if globals.quiet < 2:
                        print "MPNG::conflict -- A200 --> A800"
                    state = MPNG.a800

                if isr:
                    rpt = self.is_repeated_state( beg_cnt, net_cnt )
                    if rpt:
                        ndm_dup += 1
                        state = MPNG.a800 # dup, skip lower levels # 03/16 10:30
                        if globals.quiet < 2:
                            print "MPNG::ndm_dup -- A200 --> A800"
                            state = MPNG.a800

                if t600:
                    cur = self._o_ndm.cur( )
                    if rpt:
                        print "%2d %s %6s r %s" % ( self._level, "" * self._level, cur, self._o_ndm.mul( ) )
                    else:
                        print "%2d %s %6s   %s" % ( self._level, "" * self._level, cur, self._o_ndm.num( ) )
                        print "%2s %s %6s   %s" % ( ""         , "" * self._level, "" , self._o_ndm.den( ) )
                        print "%2s %s %6s   %s" % ( ""         , "" * self._level, "" , self._o_ndm.mul( ) )

            if max != 0 and beg_cnt >= max:
                break

            state = state( self )

            if globals.printState:
                sys.stdout.flush()

        return ( beg_cnt, net_cnt, ndm_dup, ndm_con )

    def a200( self ):

        if globals.printState:
            print self._indent, "MPNG.a200..."

        # A23400

        # A200 - Reduce Num and Den lists

        ndm = self._o_ndm.copy()
        ndm.reduce( )

        ( ndm, updated_max ) = ndm.quick_reduce( self._reduce_max )

        if self._reduce_max < updated_max:
            # numerator duplicates value in denominator
            # raise reduction threshold & restart the cycle
            self._reduce_max = updated_max
            return MPNG.a200

#       self._o_ndm = ndm.copy( )
        prv = self._o_ndm.copy( )

        if globals.testDebug:
            print self._indent, "MPNG.a200: lvl: %s" % self._level
            print "%s" % self._o_ndm.mul( )

        self._o_ndm.verifyUnusedExponentSum( )

        num = self._o_ndm.num( )
        last = num.last( )
        if not last:
            # 07/15/07 - nothing to do
            pydb.debugger( )
            print "lvl %d, last %s" % ( self._level + 1, last )
            print "prv %s" % prv
            print "ndm %s" % self._o_ndm

        self._o_ndm.rdr( last )
        self._n_rdr = last
        
        #   "A300:"
        #   "Increment level and check for multi-perfect."
        #   "If one found, save it."

        self._o_fs.append( self._o_ndm )

        self._n_cur = self._o_ndm.cur( )
        stage = MPNG_STAGE( self._o_ndm, self._n_cur, self._n_rdr )
        self._n_fs.append( stage )

        self._level += 1
        self._indent = "  " * self._level

        if len( self._o_ndm.den( ) ) == 0 and not self._o_ndm.isInteger( ):
            print "zero %s" % self._o_ndm
            self.debug( )

        if self._o_ndm.isInteger( ):

            if self._owner:
                self._owner.saveMultiperfect( self._o_ndm )
            else:
                print "%s -> %s -> %s" % ( self._seed, self._o_ndm.order( ), self._o_ndm.mul( ) )

        #   "A400:"
        #   "Scan Num and Mul for primes in Num, but not in Mul,"
        #   "adding up their exponents."

        sum = self._o_ndm.sum( )
#       assert_sum( self._o_ndm, self._n_ndm )

        #   "A401"

        if sum == 0:
            return MPNG.a800
        else:
            return MPNG.a500	# NEXT

    def a500( self ):

        if globals.printState:
            print self._indent, "MPNG.a500..."

        #   "Find avail prime (in Num and not in Mul)"
        #   "to 701 if cur already used (flag set)"
        #   "to 700 if cur already used (in Mul)"
        #   "to 600 if one found"

        rdr = self._o_ndm.rdr( )
        if rdr.used( ):
            if globals.printState and globals.quiet < 2:
                print self._indent, "MPNG.a500 -> a701 -- 1 -- %s" % rdr.prime( )
            return MPNG.a701 # NEXT
        self._o_ndm.cur( rdr )
        self._n_cur = rdr

        if not self._o_ndm.mul( ).detectPrime( rdr.prime( )):
            return MPNG.a600 # NEXT
        else:
            rdr.used( True )
            if globals.printState and globals.quiet < 2:
                print self._indent, "MPNG.a500 -> a701 -- 3 -- %s" % rdr.asUString( )
            return MPNG.a701 # NEXT
        
    def a600( self ):
        """
        Add new factor to Num, Den, and Mul lists.
        """

        cur = self._o_ndm.cur( )

        if cur.power( ) < 0:
            pydb.debugger( )

        self._o_ndm = self._o_ndm * cur

        if globals.ndm_flag and globals.qvs_flag > 1:
            print "        %s" % cur

        if globals.printState:
            print self._indent, "MPNG.a600:  %s -> %s" % ( cur, self._o_ndm.mul( ).asUString( ) )

        return MPNG.a200

    def a700( self ):

        if globals.printState:
            print self._indent, "MPNG.a700..."

        rdr = self._o_ndm.rdr( )
        if rdr.decrementPower( ):
            # process lesser power
#           if globals.printState:
#               print self._indent, "MPNG.a700 -> a600 -- r: %s" % rdr.asUString( )
            return MPNG.a600 # NEXT
        else:
            # mark as used
            rdr.used( True )
#           if globals.printState:
#               print self._indent, "MPNG.a700 -> a701 -- r: %s" % rdr.asUString( )
            return MPNG.a701 # NEXT

    def a701( self ):

        if globals.printState:
            print self._indent, "MPNG.a701..."

        lst = self._o_ndm.selectNumLessThanReader_old( )
        if len( lst ) != 0:
            rdr = self._o_ndm.rdr( lst.last( ) )
            self._n_rdr = lst.last( )

            if self._n_rdr >= self._o_ndm.num( ).first( ):
#               if globals.printState:
#                   print self._indent, "MPNG.a701 -> a500 -- r: %s" % rdr.asUString( )
                return MPNG.a500

#       if globals.printState:
#           print self._indent, "MPNG.a701 -> a800..."

        return MPNG.a800

    def a800( self ):
        """
        Back-up level and check for more to do (cur_s != 0).
        If so, decrement cur_q and cur_s.
        """

        if globals.printState:
            print self._indent, "MPNG.a800..."

        if self._level == 0:
            return None

        result_state = None

        while result_state == None:

            done = self._level == 0
            self._indent = ""

            while not done:

                # remember ndm for is_repeated_state
                self.update_states_seen( )

                # decrement level
                self._level -= 1
                self._indent = "  " * self._level

                if self._level == 0:
                    return None    # FINI

                # pop stack
                self._o_fs.pop( )
                self._o_ndm = self._o_fs[ self._level - 1 ]

                # pop stack
                self._n_fs.pop( )
                stage = self._n_fs[ self._level - 1 ]
                ( self._n_ndm, self._n_cur, self._n_rdr ) = stage.value ( )
#               assert_ndm( self._n_ndm, self._o_ndm )
#               assert_cur( self._n_cur, self._o_cur )
#               assert_rdr( self._n_rdr, self._o_ndm.rdr( ) )

                sum = self._o_ndm.sum( )
#               assert_sum( self._o_ndm, self._n_ndm )

                # 071707
                # done = sum != 0
                # 0 indicates all done
                # 1 indicates only value has been tried
                done = sum > 1

            q = self._o_ndm.power( )
            if q == 1:
                continue

            # ndm::decrementPower has supportive logic
            self._o_ndm.decrementPower( )

            if self._o_ndm.power( ) != 0:
#               if globals.printState:
#                  print self._indent, "MPNG.a800 -> MPNG.a600"
                if globals.testDebug: self.debug( )
                result_state = MPNG.a600

            if self._o_ndm.sum( ) != 0:
#               if globals.printState:
#                   print self._indent, "MPNG.a800 -> MPNG.a700"
                if globals.testDebug: self.debug( )
                result_state = MPNG.a700

        return result_state

    def testConflict( self ):

        if  globals.testConflict and self._o_ndm.conflict( ):
            if globals.quiet < 2:
                print "MPNG::tc -- %s" % self._o_ndm.mul( )
            return True
        else:
            return False

    def is_repeated_state( self, beg_cnt, net_cnt ):

        rpt = False

        if globals.ndm_flag:
            key = self._o_ndm.asString( ) # 03/16 13:45
            if key in self.state_seen:
                msg = "d"
                rpt = True
            else:
                msg = "n"

            if globals.qvs_flag > 1:
                print "%5s %s" % ( msg, key )

        if globals.testDebug and globals.quiet < 2:
            print "MPNG.next: b %d, t %d" % ( beg_cnt, net_cnt ),
            print "%s" % self._o_ndm.mul( )
            self.debug( )

        if globals.printCycle or globals.testDebug:

            if globals.printState:
                print self._o_ndm.asUString( )
            if self._o_ndm.cur( ).power( ) < 0:
                print "   power: %d" % self._o_ndm.cur( ).power( )

        return rpt

    def update_states_seen( self ):

        key = self._o_ndm.asString( ) # 03/16 13:45
        self.state_seen.add( key )    # 03/16 11:00

    def save( self, filename ):
        if self._owner:
            self._owner.save( filename )

def do_letter_arg( arg, mpng, factors ):
    
    global verbose

    if arg == "i":
        factors = Factors.Factors( None )

    if arg == "l":
        mpng.load( )

    if arg == "m":
        testMode ^= True

    if arg == "d":
        testDebug += 1

    if arg == "q":
        globals.quiet += 1

    if arg == "v":
        verbose += 1

    if arg == "r":
        ndm = NumDenMul.new( factors )
        globals.printCycle = False
        globals.quiet += 2

        ( count, beg_cnt_n, net_cnt_n, ndm_dup_n ) = mpng.search( ndm, factors )

        if count > 100:
            print "%d cycles" % count

        factors = Factors.Factors( None )

    return factors

def do_arg_with_incr( label, var, arg ):
    if len( arg ) == 2:
        var += 1
    else:
        var = int( arg[ 3: ] )
    if label:
        print "%s: %d" % ( label, var )
    return var

def main( ):

    beg = time.time( )

    globals.quiet = 0

    globals.verbose = 0
    globals.ndm_flag = 1	# NZ if state_seen set used to trim search tree
    globals.qvs_flag = 0

    global filename
    filename = RSLTS_txt

    globals.printDate   = False # True
    globals.printResult = False # True
    globals.printState  = False # True
    globals.testMode    = False		# True if comparing old/new tree traversal

    globals.printCycle  = False # True
    globals.testConflict= True  # False	# True if ndm.conflict() used to trim search tree
    globals.testDebug   = False # True

    mpng = MPNG( None )

    global isr, t600, a600

    if len( sys.argv ) == 1:

        PrintVersion( )

        globals.quiet += 2

        mpng.load( )
        factor = Number.Number( 2, 1 )
        seed = Factors.Factors( factor )

        mpng.search( 0, seed)

    else:

        factors = Factors.Factors( None )
        for arg in sys.argv[1:]:

            if arg == "new":  continue
            if arg == "old":  continue

            if len( arg ) == 1:
                factors = do_letter_arg( arg, mpng, factors )
                continue

            if arg[ 0:4 ] == "isr:":
                isr = arg[ 4: ] == 't'
                continue

            if arg[ 0:5 ] == "a600:":
                a600 = arg[ 5: ] == 't'
                print "a600 = %s" % a600
                continue

            if arg[ 0:5 ] == "t600:":
                t600 = arg[ 5: ] == 't'
                print "t600 = %s" % t600
                continue

            if arg[ 0:7 ] == "profile" or arg[ 0:6 ] == "pstats":
                profile( arg )
                continue

            arg12 = arg[ 0:2 ]

            if arg[ 0 ] == "f":
                # fq - q is 2's exponent
                q = arg[ 1: ]
                two = Number.Number( 2, q )
                verbose = 2
                triples = mpng.generate_start_triples( two )
                keys = triples.keys( )
                keys.sort( )
                for key in keys:
                    print triples[ key ],
                print
                continue

            if arg[ 0 ].isdigit( ):
                # "p" or "p^q"
                num = Number.FromString( arg )
                factors.add( num )
                continue

            if arg12 == "pd":
                printDate = do_arg_with_incr( "printDate", printDate, arg )
                continue

            if arg12 == "pr":

                if arg == "pro":
                    # PrintResultsByOrder
                    globals.PrintResultsByOrder = True
                    continue

                if arg == "prv":
                    # PrintResultsByValue
                    globals.PrintResultsByOrder = False
                    continue

                printResult = do_arg_with_incr( "printResult", printResult, arg )
                continue

            if arg12 == "ps":
                printState = do_arg_with_incr( None, printState, arg )
                continue

            if arg12 == "pc":
                printCycle = do_arg_with_incr( "printCycle", printCycle, arg )
                continue

            if arg12 == "tc":
                testConflict ^= True
                continue

            if arg12 == "td":
                testDebug = do_arg_with_incr( "testDebug", testDebug, arg )
                continue

            if arg[0] == "r" or arg[0] == "R":
                if len( arg ) == 1:
                    arg = RSLTS_txt

                if len( arg ) != 2:
                    filename = arg
                    continue

                if arg[0:4] == "rslts" or  arg[0:4] == "RSLTS":
                    continue

                if arg == "rd1":
                    PrintVersion( )

                    arg = "r1"
#                   ndm_flag = 1
                    globals.quiet += 2
#                   printCycle   = False
#                   testConflict = True
#                   testDebug    = False

                if arg == "rw":
                    mpng.load( filename )
                    mpng.save( filename + ".new" )
                    return

            if arg[0] == "c":
                print "         %s" % MPNG.instance
                print "MPNG.fc:  %s" % MPNG.instance._fc  #FactorCache( )
                print "MPNG.sfc: %s" % MPNG.instance._sfc #SigmaFactorCache( )
                
                print "MPNG.FC:  %s" % FactorCache.instance( )
                print "MPNG.SFC: %s" % SigmaFactorCache.instance( )
                print "MPNG.RC:  %s" % ResultCache.instance( )
                continue

            if arg == "ndm":
                ndm_flag += 1
                continue

            if arg == "qvs":
                qvs_flag += 1
                continue

            if arg[ 0 ] == "p":
                profile( arg )
                continue

            if arg == "x":
                # print results cache
                break

            if arg == "X":
                return

            print "unknown option - %s" % arg
            continue

        if len( factors.factors( ) ) != 0:
            ndm = NumDenMul.new( factors )
            printCycle = False
            globals.quiet += 2
#           testDebug = True
            ( count, net_cnt, ndm_dup, ndm_con ) = mpng.search( 0, factors, ndm )
            print "%d cycles" % count

        fin = time.time( )
        tim = fin - beg
        print "time: %4.2f" % tim

        print
        print globals.ResultCache.counts( )
        print

def profile_2_52_3_20_5_12_7_8( ):

    mpng = MPNG( None )
    factors = Factors.Factors( None )

    num = Number.FromString( "2^52" )
    factors.add( num )
    num = Number.FromString( "3^20" )
    factors.add( num )
    num = Number.FromString( "5^12" )
    factors.add( num )
    num = Number.FromString( "7^8" )
    factors.add( num )

    ndm = NumDenMul.new( factors )

    printCycle = False
    globals.quiet += 2

    cycleMax = 0
    count = mpng.cycle( ndm, cycleMax )

    print "%d cycles" % count

    return

def profile_2_3( ):

    mpng = MPNG( None )
    factors = Factors.Factors( None )

    num = Number.FromString( "2^3" )
    factors.add( num )

    ndm = NumDenMul.new( factors )

    printCycle = False
    globals.quiet += 2

    cycleMax = 0
    count = mpng.cycle( ndm, cycleMax )

    print "%d cycles" % count

    return

def profile_2_23( ):

    mpng = MPNG( None )
    factors = Factors.Factors( None )

    num = Number.FromString( "2^23" )
    factors.add( num )

    ndm = NumDenMul.new( factors )

    printCycle = False
    globals.quiet += 2

    cycleMax = 0
    count = mpng.cycle( ndm, cycleMax )

    print "%d cycles" % count

    return

def hotshot_stats( ):
    import hotshot, hotshot.stats, test.pystone

    prof = hotshot.Profile("stones.prof")
    benchtime, stones = prof.runcall(test.pystone.pystones)
    prof.close()

    stats = hotshot.stats.load("stones.prof")
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    stats.print_stats(20)

def profile( arg ):

    import profile

    globals.quiet = 1

    method = profile_2_3
    method = profile_2_52_3_20_5_12_7_8
    method = profile_2_23
    name = method.__name__

    if arg == "pr":
        profile.run( method )
        return

    if arg[ 0:7 ] == "profile":
        if len(arg) > 7 and  arg[ 7 ] == "_":
            name = arg

#       profile.run( method, 'profile.prof')

        import hotshot
        
        filename = method.__name__ + ".prof"
        prof = hotshot.Profile( filename )
        prof.runcall(method)
        prof.close()

        return

    if arg[ 0:6 ] == "pstats":
        if len(arg) > 6 and arg[ 6 ] == "_":
            name = arg

#       p = pstats.Stats('profile.prof')

        import hotshot.stats

        filename = method.__name__ + ".prof"
        stats = hotshot.stats.load( filename )
        stats.strip_dirs()
        stats.sort_stats('time', 'calls')
        stats.print_stats(20)
        
        return

def control_is_repeated_state( arg ):

    # 02/17/08 16:14 - control is_repeated_state
    #
    #   isr:t - True
    #   isr:f - False
    #
    #   isr:f - uses less memory
    #         - slightly faster
    #           f - 36.706s
    #           t - 37.898s
    #

    global isr
    isr = arg[ 4: ] == 't'
    print "2 - isr: %s %s" % ( arg[ 4: ], isr )

def assert_cur( n,  o ):
    global a600
    if a600 and o.magnitude( ) != n.magnitude( ):
        print "assert_cur: %s %s" % ( o.magnitude( ), n.magnitude( ) )
        pydb.debugger( )

    return

def assert_lst( n, o ):
    global a600
    if a600 and n != o:
        print "assert_lst: %s %s" % ( n, o )
        pydb.debugger( )

    return

def assert_ndm( n, o ):
    global a600
    if a600:
        ns = "%s" % n
        os = "%s" % o
        eq = ns == os
        if not eq:
            print "assert_ndm: %s %s" % ( n, o )
            pydb.debugger( )
        ni = n.isInteger( )
        oi = o.isInteger( )
        if ni != oi:
            print "assert_ndm: %s %s" % ( n, o )
            pydb.debugger( )

    return

def assert_pwr( n, o ):
    global a600
    if a600 and n != o:
        print "assert_pwr: %s %s" % ( n, o )
        pydb.debugger( )

    return

def assert_rdr( n, o ):
    global a600
    if a600 and n != o:
        print "assert_rdr: %s %s" % ( n, o )
        pydb.debugger( )

    return

def assert_sum( n, o ):
    global a600
    if not a600:
        return

    ns = n.sum( )
    os = o.sum( )
    if ns != os:
        print "assert_sum: %s %s, %s %s" % ( n.sum( ), n, o.sum( ), o )
        pydb.debugger( )

    return

if __name__ == '__main__':

    main( )
