#!/usr/bin/python

import globals
import pydb

import re, time, types

import Number
import Factors

verbose = False
NDM_verbose = False

def new( factorlist ):
    ndm = NumDenMul( )
    for number in factorlist.factors( ):
        if verbose:
            print ""
            print "       f %s" % number
        ndm.addDen( number )
        ndm.addMul( number )
        sgmftr = number.sgmftr( )
        for f in sgmftr.values():
            if verbose:
                print "       s %s" % f
            ndm.addNum( f )
        if verbose:
            print "%s" % ndm
        ndm.reduce( )
        if verbose:
            print "%s" % ndm
#        ndm.printState( )

    if verbose:
        print ""
        print "n: %s" % ndm

    return ndm

class NumDenMul:

    def __init__( self, num = None, den = None, mul = None  ):
        self._verbose = False
        self._cur = Number.Number( )
        if num != None:
            self._num = num
        else:
            self._num = Factors.new( None )
        if den != None:
            self._den = den
        else:
            self._den = Factors.new( None )
        if mul != None:
            self._mul = mul
        else:
            self._mul = Factors.new( None )
        self._sum = None
        self._hash = None
        if num != None and den != None:
            self.reduce( )

    def type( self ):

        return "NumDenMul"

    def __eq__( self, other ):
        if other:
            return self._mul == other._mul
        else:
            return False

    def __hash__( self ):
        if self._hash == None:
            self._hash = self._mul.__hash__( )
#       if self._hash != self._mul.__hash__( ):
#           print "NumDenMul::__hash__() failure"
#           pydb.debugger( )
        return self._hash

    def copy( self ):
        # answer a copy of self

        new = NumDenMul( )
        new._num = self._num.copy( )
        new._den = self._den.copy( )
        new._mul = self._mul.copy( )
        new._cur = self._cur.copy( )

        new._sum = None
        new._hash = None

        return new

    def isInteger( self ):
        """
        Answer true if instance variable den is empty.
        """

        return len( self._den ) == 0

    def isPrime( self ):
        """
        Answer true if instance variable den is empty.
        """

        return len( self._den ) == 0

    def cur( self, aFactor = None ):
        """
        Get/Set the instance variable cur.
        """

        if aFactor != None:
            self._sum = None
            self._cur = aFactor.copy( )

        return self._cur

    def decrementPower( self ):
        """
        Decrement the instance variable power.
        ... and the sum of unused powers.
        """

        p = self._cur.prime( )
        q = self._cur.power( )

#       if p == 0 or q == 0 or q == 1:
#           pydb.debugger( )

        if self._sum != self.sumUnusedExponents( ):
            print "NDM::dp_3 cur %s, sum %s, sue %s" % ( self._cur, self._sum, self.sumUnusedExponents( ) )

        self._sum -= 1
        self._cur.power( q - 1 )

#       if self._sum != self.sumUnusedExponents( ):
#           self._verbose = True
#           pydb.debugger( )
#           self._verbose = False

#       if self._cur.power( ) < 1:
#           pydb.debugger( )

#       if self._sum < 0 or q < 1:
#           pydb.debugger( )
#           print "NumDenMul::dp_8  sum - %d, cur - %s ( %d -> %d )" % ( self._sum, cur, q, q - 1 )
#           print "*************************"

        return self._cur

    def __mul__( self, aNumber ):
        """
        Add aNumber to den and mul.
        Add sgmftr(aNumber) to num.
        Reduce.
        """

        if aNumber.power( ) <= 0:
            print "NumDenMul::__mul__ %s * %s" % ( self, aNumber )
            self.debug( )

        answer = self.copy( )

        if aNumber.prime( ) == 37:
            if verbose > 0:
                print "n: %s" % aNumber
                print "s: %s" % self

        sgmftr = aNumber.sgmftr( )
        answer._num *= sgmftr
        answer._den *= aNumber
        answer._mul *= aNumber
        answer.reduce( )

#       if aNumber.prime( ) == 37:
#           revert = answer / aNumber
#           if verbose > 0:
#               print "a: %s" % answer
#               print "r: %s" % revert
#           if not revert == self:
#               print "Error"
#               pydb.debugger( )

        return answer

    def __div__( self, aNumber ):
        """
        Remove aNumber from den and mul.
        Remove sgmftr(aNumber) from num.
        Reduce.
        """

        if aNumber.power( ) <= 0:
            print "NumDenMul::__div__ %s * %s" % ( self, aNumber )
            self.debug( )

        answer = self.copy( )

        sgmftr = aNumber.sgmftr( )
        answer._num *= aNumber
        answer._den *= sgmftr
        answer._mul /= aNumber
        answer.reduce( )

        return answer

    def addDen( self, factor ):
        """
        Add a factor to instance variable den.
        """

        self._den.add( factor.copy( ) )

    def addMul( self, factor ):
        """
        Add a factor to instance variable den.
        """

        self._mul.add( factor.copy( ) )

    def addNum( self, factor ):
        """
        Add a factor to instance variable den.
        """

        self._num.add( factor.copy( ) )

    def den( self, aFactorList = None ):
        """
        Get/Set the instance variable den.
        """

        if aFactorList != None:
            self._den = aFactorList

        return self._den

    def mul( self, arg = None ):
        """
        Get/Set the instance variable mul.
        """

        if arg != None:
            if isinstance( arg, types.IntType ):
                return self._mul[ arg ]
            else:
                self._mul = arg

        return self._mul

    def num( self, aFactorList = None ):
        """
        Get/Set the instance variable num.
        """

        if aFactorList != None:
            self._num = aFactorList

        return self._num

    def power( self, anInteger = None ):
        """
        Get/Set the instance variable power.
        """

        return self._cur.power( anInteger )
   
    def prime( self, anInteger = None ):
        """
        Get/Set the instance variable prime.
        """

        return self._cur.prime( anInteger )
   
    def rdr( self, aFactor = None ):
        """
        Get/Set the instance variable rdr.
        """

        if aFactor != None:
            # must use .copy with new_rdr
            self._rdr = aFactor.copy( )

        return self._rdr
  
    def conflict( self ):
        """
        Answer True if Num and Mul overlap
        """

        #  *  ( 17^1 )
        #  n: ( 3^2 5^1 )
        #  d: ( 2^4 )
        #  m: ( 2^7 3^1 17^1 )

        ans = 1
        str = ""

        for n in self._num.values( ):
            for m in self._mul.values( ):
                if n.prime( ) == m.prime( ):
                    ans = ans * n.magnitude( )
                    if globals.quiet < 2:
                        str = "%s %s" % ( str, n )
                    break

        if ans < 10:
            return False
        else:
            if globals.quiet < 2:
                print "NDM::conflict: %s %s" % ( ans, str )
            return True

    def selectNumLessThanReader_old( self ):   #0315

        lst = self._num.selectLessThan( self._rdr )

        return lst

    def selectNumLessThanReader_new( self, rdr ):   #0315

        lst = self._num.selectLessThan( rdr )

        return lst

    def factors_multiply( self, factors ):
        
        answer = self
        for factor in factors.values():
            answer = answer * factor
        answer.reduce( )

        return answer

    def sigma_multiply( self, aNumber ):

        answer = self.copy( )
        answer._num = self._num * aNumber.sgmftr( )
        answer._den = self._den * aNumber
        answer._mul = self._mul * aNumber

        return answer

    def order( self ):

        return self._num.magnitude( )

    def sum( self ):
        """
        Get/Set the instance variable sum to anInteger.
        """

        if not self._sum:
            self._sum = self.sumUnusedExponents( )
        else:
            s1 = self._sum
            s2 = self.sumUnusedExponents( )
            if s1 != s2:
                pydb.debugger( )

        return self._sum
  
    def sumUnusedExponents( self ):
        """
        Answer the sum of the unused exponents.
        """

        answer = 0
        for factor in self._num.factors( ):
            if factor.used( ):
                continue
            if self._mul.has_key( factor.prime( ) ):
                continue
            if factor.prime( ) != self._cur.prime( ):
                answer += factor.power( )
            else:
                answer += self._cur.power( )
            if self._verbose:
                print "%s %s" % ( factor.asUString(), answer )

        return answer

    def verifyUnusedExponentSum( self ):

        s1 = self.sum( )
        s2 = self.sumUnusedExponents( )
        if s1 == s2:
            return True
        else:
            globals.PROG.debug( )
            NDM_verbose = True
            s2 = self.sumUnusedExponents( )
            NDM_verbose = False
            for factor in self._num.factors( ):
                print factor
            return False

    def trace( self, label, factors ):

        print "-- %s %d %d %d %s" % ( label, len( factors ), len( factors.numbers.keys()), len( factors.numbers.values()), factors.numbers.keys( ) )

    def asUString( self ):

        answer = ""
        answer += "  n: %s\n" % self._num.asUString( )
        answer += "  d: %s\n" % self._den.asUString( )
        answer += "  m: %s"   % self._mul.asUString( )

        return answer

    def reduce( self ):

        for nf in self._num.factors( ):
            fp = nf.prime( )
            fq = nf.power( )
            if not self._den.has_key( fp ):
                continue
            changed = True
            dq = self._den[ fp ].power( )
            if fq == dq:
                self._num.remove( fp )
                self._den.remove( fp )
            elif fq > dq:
                self._num[ fp ].power( fq - dq )
                self._den.remove( fp )
            else:
                self._num.remove( fp )
                self._den[ fp ].power( dq - fq )

        return self

    def quick_reduce( self, max ):
        # reduce numerator factors greater than 'max'
        # answer self.copy and updated max

        answer = self.copy( )
        if verbose > 0:
            print "mx: %s %s" % ( max, answer )

        while True:

            factor = answer._num.last( )
            fp = factor.prime( )
            if fp <= max:
                if verbose > 0:
                    print "fp <= max ( %d, %d )" % ( fp, max )
                break
            if not answer._mul.has_key( fp ):
                answer *= factor
                op = "*"
            else:
                answer /= factor
                op = "/"

            if verbose > 1:
                print "fp %s: %s %s" % ( op, fp, answer )

        if verbose > 0:
            print "fp: %s %s" % ( fp, answer )

        return ( answer, fp )

    def reduceLargeFactors( self, max ):
        """
        Reduce large ndm.num primes (to less than 10)
        """

        changed = True
        answer = self.copy( )

        if verbose:
            print "reduceLargeFactors( %s )" % answer

        while changed:
            changed = False
            values = answer.num( ).sortedvalues( )
            values.reverse( )
            if values[ 0 ].prime( ) < max:
                break
            print values
            for value in values:
                p = value.prime( )
                if p <= max:
                    continue
                if not answer.mul( ).hasPrime( p ):
                    answer *= value
                    if verbose:
                        print "* %s -> %s" % ( value, answer )
                    changed = True
                else:
                    break
                    mul = answer._mul / value
                    answer = new( mul )
                    if verbose:
                        print "/ %s -> %s" % ( value, answer )

        return answer

    def used( self, aBoolean = None ):
        """
        Set the instance variable used to aBoolean.
        """

        return self._cur.used( aBoolean )
            
    def __str__( self ):
        """
        Answer the receiver as a formatted string.
        """

        answer = "(\n"
        answer += "  num: %s\n" % self._num
        answer += "  den: %s\n" % self._den
        answer += "  mul: %s\n" % self._mul
        show_sum_and_reader = False
        if show_sum_and_reader:
            try:
                answer += "  sum: %s, rdr: %s\n" % ( self.sum( ), self.rdr( ) )
            except:
                answer += "  sum: %s\n" % self.sum( )
        answer += ")"

        return answer

    def asString( self ):

        answer = self._mul.asString( )
        
        return answer

    def printState( self ):
        print self
        if self.isPrime( ):
            print "multiperfect"
        else:
            print "not"
            print ""

def Test( ):
    factors = Factors.Factors( )
    factors.addString( "2^3" )
    factors.addString( "3^1" )
    factors.addString( "5^1" )
    NDM = new( factors )
    print NDM

#    print "1: %s" % NDM
#    NDM.decrementPower( )
#    print "2: %s" % NDM
#    NDM.decrementPower( )
#    print "3: %s" % NDM

if __name__ == '__main__':

    import sys
    from mpz import *

    verbose = 0

    if len( sys.argv ) == 1:
        Test( )
    else:
        ndm = NumDenMul( )
        beg = time.time( )
        quiet = False

        globals.quiet = 1
        globals.save = False

        for arg in sys.argv[ 1: ]:
            if arg == "v":
                verbose += 1
                continue

            if arg[ 0:2 ] == "qr":
                #NumDenMul.py qr:2^20:17
                #NumDenMul.py qr:2^17:17
                ( p, q, m ) = re.split( "[=:.,\^]", arg[ 3: ] )
                p = mpz( p )
                q = mpz( q )
                ndm = NumDenMul( )
                num = Number.Number( p, q )
                ndm = ndm.sigma_multiply( num )
                ( ndm1, n ) = ndm.quick_reduce( m )
                ndm2 = ndm.reduceLargeFactors( m )
                if ndm1 == ndm2:
                    print "%s - OK - %s" % ( arg, ndm2 )
                else:
                    print "%s - Error" % arg
                    pydb.debugger( )

                quiet = True
                continue

            if arg[ 0 ].isdigit( ):
                ( p, q ) = re.split( "\^", arg )
                p = mpz( p )
                q = mpz( q )

                mul = ndm.mul()
                if mul.has_key( p ):
                    print "dup key - %s^%s" % ( p, q )
                    quiet = True
                else:
                    num = Number.Number( p, q )
                    ndm = ndm.sigma_multiply( num )
                    ndm.reduce( )
                        
                continue

            print "unknown arg: %s" % arg

        if not quiet and len( ndm._mul ) != 0:
            if len( ndm._den ) == 0:
                print "*** multiperfect ***"
            print ndm

        fin = time.time( )
        if beg - fin > 1:
            print " time: %4.2f" % ( fin - beg )
