#!/usr/bin/python

import globals
import pydb
import re
from mpz import *

import Factors
import SigmaFactor

class Number:

    """
    A holder for a prime (p) and its exponent (q).
    """

    def __init__( self, prime=0, power=0, used=False, str=None ):
        self._p = mpz( prime )
        self._q = mpz( power )
        self._u = used
        self._s = str

    def type( self ):

        return "Number"

    def new( self, prime, power ):
        n = self.___init__( self, prime, power )

    def copy( self ):
        # answer a copy of self
        return Number( self._p, self._q, self._u, self._s )

    def __cmp__( self, other ):
        sp = self.prime( )
        op = other.prime( )
        sq = self.power( )
        oq = other.power( )

        if sp != op:
            return sp - op
        else:
            return sq - oq

    def __hash__( self ):
        an = self._p * self._q
        an = int( an % 1000000 )
        return an

    def __str__( self ):
        """
        Format as string
        """

        return self.asString( )

    def __repr__( self ):
        """
        Format as string
        """

        return self.asString( )

    def asString( self ):
        """
        Format as string
        """

        if self._s == None:
            self._s =  "%s^%s" % ( self._p, self._q )

        return self._s

    def asUString( self ):
        ans = self.asString( )
        if self._u:
            ans += ".u"
        else:
            ans += " "
        return ans

    def decrementPower( self ):
        if self._q <= 1:
            return False
        else:
            self._q -= 1
            self._s = None
            return True

    def isPrime( self ):
        if self._q == 1:
            return True
        else:
            return False

    def magnitude( self ):
        """
        Answer the instance variable q.
        """

        return self._p ** self._q

    def power( self, anInteger = 0 ):
        """
        Get/Set the instance variable q.
        """

        if anInteger:
            self._q = anInteger
            self._s = None

        return self._q

    def prime( self, anInteger = None ):
        """
        Get/Set the instance variable p.
        """

        if anInteger:
            self._p = anInteger
            self._s = None

        return self._p

    def used( self, aBoolean = None ):
        """
        Get/Set the instance variable u.
        """

        if aBoolean != None:
            self._u = aBoolean

        return self._u

    def sgmftr( self ):
        """
        Answer the sum of the factors of prime^power.
        """

        if self._q <= 0:
            print "Number::sgmftr( %d, %d )" % ( self._p, self._q )
            pydb.debugger( )

        sf = SigmaFactor.SigmaFactor( self._p, self._q )
        an = sf.value( )

        return an

    def mersenne( self ):
        return None
        ##################################################
########        if self._p != 2:
########            return None
########        q = self._q + 1
########        f = Factors.Factors( q )
########        if f.isPrime( ):
########            print "prime"
########        else:
########            if Factors.isMersennePrime( q ):
########                v = 2 ** ( self._q + 1 ) - 1
########                n = Number.Number( v, 1 )
########                globals.FactorCache
#########            if self.LucasLehmerTest( 2 ** ( self._q + 1 ) - 1 ):
        ##################################################

    def __eq__( self, other ):
        if other == None:
            return False
        if self._p != other._p or self._q != other._q:
            return False
        return True
        
    def __ne__( self, other ):
        if other == None:
            return True
        if self._p != other._p or self._q != other._q:
            return True
        return False
        
    def __lt__( self, other ):
        if self._p < other._p or \
           self._p == other._p and self._q < other._q:
            return True
        else:
            return False
        
    def __le__( self, other ):
        if self._p < other._p or \
           self._p == other._p and self._q <= other._q:
            return True
        else:
            return False
        
    def __gt__( self, other ):
        if self._p > other._p or \
           self._p == other._p and self._q > other._q:
            return True
        else:
            return False
        
    def __ge__( self, other ):
        if self._p > other._p or \
           self._p == other._p and self._q >= other._q:
            return True
        else:
            return False
        
def FromString( aString ):

    if aString.find( "^" ) < 0:
        p = aString
        q = "1"
    else:
        ( p, q ) = re.split( "[\^]", aString )

    result = FromStringPair( p, q )

    return result

def FromStringPair( p, q ):

    p = mpz( p )
    q = mpz( q )

    result = Number( p, q )

    return result

def Test( ):
    print "Number.Test:"
    for p in ( 5,7 ):
        for q in ( 1,2,3 ):
            n = Number( p, q )
#           print "  n:   %s" % n
            print "  p^q: %s^%s %3s %s" % ( n.prime( ), n.power( ), n.magnitude( ), n.sgmftr( ) )

def Test_Compare( ):
    print "Test_Compare:"
    n1 = Number( 5, 3)
    n2 = Number( 5, 4)
    if n1 < n2:
        print " %s <  %s" % ( n1, n2 )
    else:
        print " %s >= %s" % ( n1, n2 )

    if n1 > n2:
        print " %s >  %s" % ( n1, n2 )
    else:
        print " %s <= %s" % ( n1, n2 )

def Test_Powers( ):
    exponents = ( 1, 2, 5, 10, 20, 50 )
    exponents = ( 1, 2, 5,
                  10, 20, 50,
                  100, 200, 500,
                  1000, 2000, 5000,
                  10000, 100000 )
    for q in exponents:
        v = 2 ** q
        print "%3d %d %d" % ( q, len( "%s" % v ), v )

if __name__ == '__main__':

    Test( )
    Test_Compare( )
    Test_Powers( )
