#!/usr/bin/python

import globals
from mpz import *

import Wheel

import Number
import Factors

wheel = None
wheel_array = None

class PrimeGenerator:

    def __init__( self, order ):

        global wheel, wheel_array

        self._index = 0
        self._value = 1
        self._cycle = 0

        if not wheel:
            wheel = Wheel.Wheel( order )
            wheel_array = [ ]
            for item in wheel.wheel( ):
                wheel_array.append( mpz( item ) )

        self._order = order + 1
        self._size  = len( wheel_array )

        return

    def cycle( self ):
        return self._cycle

    def NextPrime( self ):
        if self._index >= self._size:
#           self._CheckStatus( )
            self._index = self._order
        self._value += wheel_array[ self._index ]
        self._index += 1
        return self._value

class PrimeBatchGenerator( PrimeGenerator ):

    def __init__( self, order ):

        PrimeGenerator.__init__( self, order )
        self._spokes = wheel_array[ 0: self._order ]
        self._wheel  = wheel_array[ self._order : ]
        self._result = self._spokes

        return

    def primes( self ):
        result = self._result
        self._cycle += 1
        self._result = self._wheel
        return result

    def gmpy( self, list ):
        array = [ ]
        for item in list:
            array.append( mpz( item ) )
        return array

    def wheel( self ):
        self._cycle += 1
        self._result = self._wheel

    def print_mpz( self, list ):
        print "[",
        sep =" "
        for item in list:
            print "%s%s" % ( sep, item ),
            sep = " ,"
        print "]"
            
def PrimeTest( PrimeCount, MaxNumber ):
    # Test routine.

    import PrimeGenerator30
    import PrimeGenerator210

    count = PrimeCount
    print "PrimeGenerator.PrimeTest: %d" % count

    # PrimeTest: 25
    #   1: [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25]
    #   6: [2, 3, 5, 7, 11, 13, 17, 19, 23, 25, 29, 31, 35, 37, 41, 43, 47, 49, 53, 55, 59, 61, 65, 67]
    # 210: [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89]
    print "     1: %s" % PrimeGenerator( ).PrimeTest( count )
    print "     6: %s" % PrimeGenerator30.PrimeGenerator30( ).PrimeTest( count )
    print "   210: %s" % PrimeGenerator210.PrimeGenerator210( ).PrimeTest( count )

    # FactorTest:     100     25 primes,      0 non-primes,     25 total
    # FactorTest:     200     46 primes,      4 non-primes,     50 total
    # FactorTest:     500     95 primes,     23 non-primes,    118 total
    # FactorTest:    1000    168 primes,     63 non-primes,    231 total
    # FactorTest:   10000   1229 primes,   1059 non-primes,   2288 total
    # FactorTest:  100000   9592 primes,  13268 non-primes,  22860 total
    # FactorTest: 1000000  78498 primes, 150076 non-primes, 228574 total
    for number in ( 100, 200, 500, 1000, 10000, 100000, 1000000 ):
        if number <= MaxNumber:
            FactorTest( PrimeGenerator210, PrimeGenerator210, number, False, False )

def FactorTest( clPG1, clPG2, anInteger, fTime, fVerbose):
    # Test routine.
    # Compute the number of pseudo-primes in the interval.
    # Display the number of primes, non-primes, and pseudo-primes.

    print "PrimeGenerator.FactorTest: %d" % anInteger

    if fTime:
        beg = time.time( )

    cc = 0
    pc = 0
    g1 = PrimeGenerator.PrimeGenerator( order )
    g2 = PrimeGenerator.PrimeGenerator( order )
    while  True:
        p = g1.NextPrime( )
        if p >= anInteger:
            break

        f = Factors.Factors( p )
        f = f.factor( p, g2 )

        if len( f ) == 1 and f.values()[0].isPrime:
            pc += 1
            if fVerbose:
                print "  %d %s - prime" % ( p, f )
        else:
            cc += 1
            if fVerbose:
                print "  %d %s - not prime" % ( p, f )

    print "  %d primes, %d non-primes, %d total" % ( pc, cc, pc + cc )
    if fTime:
        fin = time.time( )
        print "%4.3f" % ( fin - beg )

    sys.stdout.flush( )

def SgmftrTest( ):

    print "PrimeGenerator.SgmftrTest:"

    for p in range( 5,9,2 ):
        for q in range( 1,4 ):
            n = Number.Number( p, q)
            print "  %s %s" % ( n, n.sgmftr( ) )

if __name__ == '__main__':

    hash = { }
    hash[   6 ] = 2
    hash[  30 ] = 3
    hash[ 210 ] = 4

    for argv in sys.argv:

        if argv[ 0 ].isdigit( ):
            arg = int( argv )
            PrimeGenerator( arg )
            PrimeBatchGenerator( arg )

        if argv == "qf":
            for cl1 in ( 30,  210 ):
                for cl2 in ( 30, 210 ):
                    FactorTest( hash[ cl1 ], hash[ cl2 ], 100, True, False )

        #  30:  9592 primes, 23742 non-primes, 33334 total
        # 210:  9592 primes, 13268 non-primes, 22860 total
        #
        # 10.589  6.133    10.484  6.050    10.455  6.096
        #  8.517  4.815     8.550  4.820     8.520  4.797
        #
        if argv == "f":
            for cl1 in ( PrimeGenerator30, PrimeGenerator210 ):
                for cl2 in ( PrimeGenerator30, PrimeGenerator210 ):
                    FactorTest( cl1, cl2, 100000, False, False )

        if argv == "f_30_30":
            FactorTest( hash[ 30 ], hash[ 30 ], 100000, True )
        if argv == "f_30_210":
            FactorTest( hash[ 30 ], hash[ 210 ], 100000, True )
        if argv == "f_210_30":
            FactorTest( hash[ 210 ], hash[ 30 ], 100000, True )
        if argv == "f_210_210":
            FactorTest( hash[ 210 ], hash[ 210 ], 100000, True )

        if argv == "s":
            SgmftrTest( )

        if argv[0] == "p":
            if len( argv ) == 1:
                PrimeTest( 25, 0 )
            else:
                count = int( argv[1:] )
                print count
                PrimeTest( 100, count )
