#!/usr/bin/python

import globals
import pydb

import re, types

import Cache
import Factors
import Result
import ResultCounter
import TimeStamp

class ResultAndSeedSet:
    '''
    A result (multiperfect number) and the set of seeds that generate it
    '''

    def __init__( self, result, seed ):
        self._result = result
        self._seeds = set( )
        if seed:
            self._seeds.add( seed )

    def type( self ):

        return "ResultAndSeedSet"

    def __cmp__( self, other ):

        return self._result.__cmp__( other._result )

    def __eq__( self, other ):

        if other == None:
            return False

        pydb.debugger( )

        if self._result == other._result:
            return True
        else:
            return False

    def add( self, seed ):

        # set Factors._minimal
        seed.minimal( True )

#  'intersection':                <method 'intersection' of 'set' objects>
#  'intersection_update':         <method 'intersection_update' of 'set' objects>
#  'isdisjoint':                  <method 'isdisjoint' of 'set' objects>
#  'issubset':                    <method 'issubset' of 'set' objects>
#  'issuperset':                  <method 'issuperset' of 'set' objects>
#  'pop':                         <method 'pop' of 'set' objects>
#  'remove':                      <method 'remove' of 'set' objects>
#  'symmetric_difference':        <method 'symmetric_difference' of 'set' objects>
#  'symmetric_difference_update': <method 'symmetric_difference_update' of 'set' objects>
#  'union':                       <method 'union' of 'set' objects>
#  'update':                      <method 'update' of 'set' objects>

        for factors in self._seeds:
#            for factor in factors.values( ):
#                if factor == seed:
#                    continue
                if factors.values().issubset( seed.values( ) ):
#               if factor.includes( seed ):
                    pydb.debugger( )
                    factors.minimal( False )
#                    print "temp: %s" % temp
                if seed.values().issubset( factors.values( ) ):
#               if seed.includes( temp ):
                    pydb.debugger( )
                    seed.minimal( False )
#                    print "seed: %s" % seed

        self._seeds.add( seed )

        return self

    def result( self ):
        return self._result

    def seeds( self ):
        return self._seeds

class ResultCache( Cache.Cache ):

    """
    A cache of results stored as a dictionary of dictionaries.
    Outer key is the multiperfect's order.
    Inner key is the multiperfect's factor list.
    Inner value is the multiperfect number (a Result object).
    """

    def __init__( self, saver ):
        Cache.Cache.__init__( self, saver )
        self._counter = 0
        self._counts = ResultCounter.ResultCounter( )
        return

    def type( self ):

        return "ResultCache"

    def __str__( self ):

        if globals.PrintResultsByOrder:
            # by multiperfect's order ( 2, 3, ... )
            answer = self.asStringOrderedByOrder( )
        else:
            # by multiperfect's factors ( 2^N, 3^N, ... )
            answer = self.asStringOrderedByFactors( )

        if answer == "":
            answer = "<ResultCache.ResultCache instance>"

        return answer

    def asStringOrderedByOrder( self ):

        quiet = globals.quiet
        globals.quiet = False

        blank = ""
        answer = ""

        for order in self.sortedkeys( ):
            answer += blank
            blank   = "\n"

            values = self._dict.sortedvalues( order )

            for rco in values:
                result = rco.result( )
                seeds  = rco.seeds( )
                answer += "%s\n" % result
                answer += "%s\n" % self.printSeeds( seeds )

        globals.quiet = quiet

        return answer

    def asStringOrderedByFactors( self ):

        quiet = globals.quiet
        globals.quiet = False
        
        answer = ""
        rco_list = [ ]

        for order in self.sortedkeys( ):

            results_dict = self._dict[ order ]
            for rco in results_dict.values( ):
                rco_list.append( rco )

        rco_list.sort( )

        for rco in rco_list:
            result = rco.result( )
            seeds  = rco.seeds( )
            answer += "%s\n" % result
            answer += "%s"   % self.printSeeds( seeds )

        globals.quiet = quiet

        return answer

    def printSeeds( self, seeds ):

        answer = ""
        list = [ ]
        for seed in seeds:
            list.append( seed )
        list.sort( )
        for seed in list:
            if globals.quiet:
                answer += "%s\n" % seed
            else:
                answer += "%26s %s\n" % ( " ", seed )

        return answer

    def add( self, anObject, seed = None ):

        if isinstance( anObject, types.StringType ):
            string = anObject
            answer = self.addString( string )
        else:
            answer = self.addResult( anObject, seed )

        self._counter += 1

        return answer

    def addResult( self, result, seed ):

        "Add a multiperfect to the result list."
        "Update new/old/dup counts."

        order = result.order( )

        answer = self.at( result )
        if answer == None:
            result = result.copy( )
            rco = ResultAndSeedSet( result, seed )
            results_dict = self.order( order )
            results_dict[ result.factors( ) ] = rco
            state = result.state( )
            if state != "old":
                self._counts.update( state )
            globals.Saver.newResult( )
            globals.Saver.doSave( True )

        else:
            result = answer.result( )
            seeds = answer.seeds( ) 

            state = result.state( )
            if state != "new":
                self._counts.update( state )

            if not seed in seeds:
                seeds.add( seed )
                globals.Saver.newResult( )

        return result

    def addString( self, line ):

        if ( len( re.split( ":", line ) ) != 1 ):

            tokens = re.split( " +", line )
            date = "%s %s" % ( tokens[ 0 ], tokens[ 1 ] )
            date = TimeStamp.TimeStamp( date )
            order = int( tokens[ 2 ] )
            power = int( tokens[ 3 ] )

            tokens = re.split( "[()]", line )
            factors = Factors.FromString( tokens[ 1 ] )

            state = "old"
            result = Result.Result( order, factors, date, state )

            if self.verify( line, order, result.factors( ) ):
                results = self.order( order )
                self._rco = ResultAndSeedSet( result, None )
                results[ factors ] = self._rco
                self._counts.update( state )

        else:

            seed = Factors.FromString( line )
            self._rco.add( seed )

        return

    def at( self, aResult ):
        "Answer the cached result (or nil)."

        order = aResult.order( )

        if not self._dict.has_key( order ):
            return None

        results = self._dict[ order ]
        factors = aResult.factors( )

        if results.has_key( factors ):
            answer = results[ factors ]
            return answer
        else:
            return None

    def order( self, value ):

        if self._dict.has_key( value ):
            results = self._dict[ value ]
        else:
            results = { } # dict
            self._dict[ value ] = results
        return results

    def xxx_matchFactor_xxxdef( self, two, delta = 0 ):

        pydb.debugger( )
        two_q = two.power( )
        results = set( )
        max = 0

        while 1 :
            two_q += delta

            for key in self._dict.keys( ):
                rco_list = self._dict[ key ].values( )
                for rco in rco_list:
                    result = rco.result( )
                    seeds  = rco.seeds( )
                    factor = result.factors( )[ 2 ]
                    factor_q = factor.power( )
                    if factor_q > max:
                        max = factor_q
                    if factor_q == two_q:
                        results.add( result )

            if delta == 0:
                break
            if len( results ) != 0:
                break

            if two_q == 0:
                return None
            if two_q > max:
                return None

        return results

    def matchFactors( self, factors, delta = 0 ):

        '''
        Find results that match the factor list.
        If lacking exact match, increase power and try again.
        '''

        results = set( )
        two = factors[ 2 ].copy( )
        
        while True :
            two.power( two.power( ) + delta )

            for key in self._dict.keys( ):
                rco_list = self._dict[ key ].values( )
                if rco_list == None:
                    break
                for rco in rco_list:
                    result = rco.result( )
                    if result.matchesFactors( factors ):
                        results.add( result )

            if delta == 0:
                break
            if len( results ) != 0:
                break

        return results

    def counts( self ):

        # Accessor for ResultCounter object

        return self._counts

    def updateStartTriples( self, triples, seed ):

        """
        find matching power of 2 in result cache
        find min & max powers of 3, 5, and 7
        """

        results_set = self.matchFactors( seed )

        keys = triples.keys( )
        for p in keys:
            for result in results_set:
                factors = result._factors
                if factors.has_key( p ):
                    factor = factors[ p ]
                    triple = triples[ p ]
                    triple.adjMinMax( factor )

        return

    def verify( self, line, order, factors ):

        mag = factors.magnitude( )
        sgm = factors.sgmftr( ).magnitude( )
        if sgm == mag * order:
            return True
        else:
            order = sgm * 1.0 / mag
            print "%s" % "ResultCache error:"
            print "  %s" % line
            print "  %s %0.2f" % ( " " * 20, order )
            pydb.debugger( )
            return False

##        if sgm == mag * order:
##            return order
##        else:
##            print "ResultCache error:"
##            print "  %s" % factors
##            print "  %s" % sgm
##            print "  %s" % ( mag * order )
##            print "  %s vs %0.2f" % ( order, sgm * 1.0 / mag )
##
##            if order == int( order ):
##                print "order is %d" % order
##            else:
##                print "order %0.2f is not an integer" % order
##
##            return order

#       factoring: 67108863
#       2^25  33554432                     33554432  67108863                     67108863
#       3^4         81                   2717908992        121                  8120172423
#       7^1          7                  19025362944          8                 64961379384
#       11^2       121                2302068916224        133               8639863458072
#       19^2       361              831046878756864        381            3291787977525432
#       127^1      127           105542953602121728        128          421348861123255296
#       683^1      683         72085837310249140224        684         8202621008306622464
#       2731^1    2731      19666421694290401951744       2732      7879560594693692571648
#       8191^1    8191  161253260097932682386735104       8192  64501310391730729546940416
