#!/usr/bin/python

import globals

import sys, time

ONE_HOUR = 60 * 60
TEN_MINUTES = 10 * 60

class Saver:

    """
    Periodically save a results file in rslts.d/rslts.MMDD
    Keep track of time since last save and whether new
    Factors, SigmaFactors, or Results have been computed.
    """

    def __init__( self, owner = None ):
#        print "Saver::ctor %s" % self
        self._owner = owner
        self.init( )

    def type( self ):

        return "Saver"

    def owner( self, anObject ):
        self._owner = anObject

    def init( self ):
#        print "Saver::init %s" % self
        self._newFactor = False
        self._newSgmftr = False
        self._newResult = False
        self._saveTime  = time.time( )

    def newFactor( self ):
        if not self._newFactor and not globals.quiet:
            print "******** newFactor ********"
        self._newFactor = True

    def newSgmftr( self ):
        if not self._newSgmftr and not globals.quiet:
            print "******** newSgmftr ********"
        self._newSgmftr = True

    def newResult( self ):
        if not self._newResult and not globals.quiet:
            print "******** newResult ********"
        self._newResult = True

    def needSave( self ):
        # save if caches updated
        elapse = time.time( ) - self._saveTime

        # no less than 10 minutes
        if elapse < TEN_MINUTES:
            return False

        # at least hourly
        if elapse > ONE_HOUR:
            return True

        # if any cache expanded
        return self._newFactor or self._newSgmftr or self._newResult

    def filename( self ):
        tm = time.localtime( time.time( ) )
        fn = time.strftime( "rslts.d/rslts.%m%d.%H%M", tm )
        return fn

    def doSave( self, force = False ):
        if force or self.needSave( ):
            filename = self.filename( )
            self._owner.save( filename )
            self.init( )
            sys.stdout.flush()
            return True

if __name__ == '__main__':
    saver = Saver( )
    globals.quiet = False
    for count in range( 0, 10):
        if ( count % 5 ) == 1:
            saver.newFactor( )
        if ( count % 5 ) == 2:
            saver.newSgmftr( )
        if ( count % 5 ) == 3:
            saver.newResult( )

        if not saver.doSave( None ):
            print "%d: %s" % ( count, False )
        else:
            print "%d: %s" % ( count, saver.filename( ) )
