#!/usr/bin/python

import globals
import pydb
import re, sys
from mpz import *

import Factors
import Number

def FromString( aString ):
    if aString.find( "^" ) < 0:
        p = int( aString )
        q = 1
    else:
        ( p, q ) = re.split( "[\^]", aString )
        p = int( p )
        q = int( q )

    f = Factors.Factors( p )
    r = Factors.new( None )

    for t in f.factors( ):
        v = SigmaFactor( t.prime( ), t.power( ) * q )
#        if not s:
#            print "sgmftr( %s, %s ) -> %s" % ( t.prime( ), q, v )
        r = r * v

    return ( p, q, r )

class SigmaFactor:
    """
    Computes sgmftr( p ** q )
    """

    def __init__( self, p = None, q = None ):
        self._p = mpz( p )
        self._q = mpz( q )
        self._v = self.sgmftr( )
        return

    def type( self ):

        return "SigmaFactor"

    def __str__( self ):
        try:
            ans = "%s" % self._v
        except:
            ans = "%s %s None" % ( self._p, self._q )
        return ans

    def value( self ):
        """
        Answers a Factors object
        """

        try:
            return self._v
        except:
            return None

    def factor( self, n ):
        return Factors.Factors( n )

    def factors( self ):
        return self._v

    def sgmftr( self ):
        """
        Answers a Factors object
        """

        if self._p * self._q > 100:
            # retrieve value from cache or generate it
            n = Number.Number( self._p, self._q )
            ans = globals.SigmaFactorCache.at( n )

            if ans != None:
                return ans

        num = self.powerMinus1( self._p, self._q + 1 )

        if self._p == 2:
            ans = num
        else:
            den = Factors.Factors( self._p - 1 )
#           if num == None or den == None:
#               pydb.debugger( )
#               print "SigmaFactor::sgmftr( %d, %d )" % ( self._p, self._q )
#               print "num: %s, den: %s" % ( num, den )
            ans = num / den

        if self._p * self._q > 100:
            # add value to cache
            if self._p == 2 and self._q == 5:
                print "pp:qq: %d %d" % ( self._p, self._q )
                n = Number.Number( self._p, self._q )
            n = Number.Number( self._p, self._q )
            globals.SigmaFactorCache.add( n, ans )

        return ans

    def powerMinus1( self, p, q ):
        """
        factors p**q-1 and answers a Factors object
        """

        if globals.verbose:
            print "%d ** %d - 1" % ( p, q )

        value = p ** q - 1

        if q == 0:
            print "SigmaFactor::powerMinus1( %d, %d )" % ( p, q )
            return

        if q == 1:
            answer = self.factor( value )
            if globals.verbose:
                print "... %s" % answer
            return answer

        if q % 2 == 0:
            # compute( p^2q - 1 ) as ( p^q - 1 ) * ( p^q + 1 )
            q = q / 2
            if q == 1:
                pri_min = Factors.Factors( p - 1 )
                pri_plu = Factors.Factors( p + 1 )
            else:
                pri_min = self.powerMinus1( p, q )
                pri_plu = self.powerPlus1(  p, q )
            answer = pri_min * pri_plu
            if globals.verbose:
                print "... %s" % answer
            return answer

        answer = Factors.Factors( )

        qfactors = Factors.Factors( q )

        if not qfactors.isPrime( ):
            factors = qfactors.factors( )
            factors.reverse( )
            for factor in factors:
                prime = factor.prime( )
                div = p ** prime - 1
                if value % div == 0:
                    dfactors = self.factor( div )
                    answer = answer * dfactors
                    value = value / div

        # possible mersenne
        if p == 2 and len( qfactors ) == 2:
            if self.isMersenne( q ):
                pydb.debugger( )
                if globals.verbose:
                    print "pm.4..%s" % value
                mersenne = [ p ** q - 1, ( p ** q - 1, 1 ) ]
                answer = answer * mersenne
                if globals.verbose:
                    print "pm.4: %s,%s -> %s" % ( p, q, answer )

        answer = answer * self.factor( value )
        if globals.verbose:
            print "... %s" % answer
        return answer

    def powerPlus1( self, p, q ):
        """
        factors p**q+1 and answers a Factors object
        """

        if globals.verbose:
            print "%d ** %d + 1" % ( p, q )

        value = p ** q + 1

        if q == 1:
            answer = self.factor( value )
            if globals.verbose:
                print "... %s" % answer
            return answer

        answer = Factors.Factors( )

        qfactors = Factors.Factors( q )

        if globals.verbose:
            print "qf - %s" % qfactors

        if len( qfactors ) == 1 and qfactors.first( ).power( ) == 1:
            # if prime
            answer = self.factor( value )
        else:
            # if not prime
            for factor in qfactors.factors( ):
                prime = factor.prime( )
                if prime == 2:
                    continue
                div = p ** prime + 1

                if value % div == 0:
                    value = value / div
                    answer = answer * self.factor( div )

            if value > 1:
                if globals.verbose:
                    print "... %s" % value
#                if value > 1000000:
#                    pydb.debugger( )
                answer *= Factors.Factors( value )

                if globals.verbose:
                    print "... %s" % answer
        return answer

    def div( self, list, value ):
        print "div: %s %s" % ( list, value )
        answer = [ ] # list
        for index in range( 0, len( list ), 2):
            prime = list[ index ]
            power = list[ index + 1]
            if prime == value:
                value = 0
                if power == 1:
                    continue
                else:
                    power -= 1
            answer += [ prime, power ]

            if globals.verbose:
                print "           %s" % answer
        return answer

    def isMersenne( self, q ):
        f = Factors.Factors( )
        return f.isLucasLehmerPrime( q )

#    ./PowerFactor.py p2 q47.50
#
#    pmcheck - OK - 2 ** 47
#                   [2351, 1, 4513, 1, 13264529, 1]
#    pmcheck - OK - 2 ** 48
#                   [7, 1, 3, 2, 65, 1, 4097, 1, 16777217, 1]
#    pmcheck - OK - 2 ** 49
#                   [127, 1, 4432676798593L, 1]

    def pmcheck( self ):
        # compute result directly
        v1 = self._p ** self._q - 1

        # compute result by factoring then multiplying
        print "pmx ..."
        pmx = self.powerMinus1( self._p, self._q )
        
        print "o: %s" % pmx.__class__

        v2 = 1
        for factor in pmx.factors( ):
            print "   %s" % factor
            v2 = v2 * factor.prime( ) ** factor.power( )

        # compare results
        if v1 != v2:
            print "pmcheck - error - %s ^ %s - 1" % ( self._p, self._q )
            print "                  %s" % v1
            print "                  %s" % v2
            print "                  %s %s" % ( v2 / v1, v2 % v1 )
            print "                  %s" % pmx

        return pmx

    def factor_simple( self, i ):

        # simplistic factoring of a number
        r = [ ] # list
        v = i
        p = 2
        while p * p <= v:
            q = 0
            while v % p == 0:
                q += 1
                v = v / p
            if q != 0:
                r += ( p, q )
            p += 1

        if v > 1:
            r += ( v, 1 )

        return r

    def sgmftr_old( self ):
        """
        Answers a Factors object
        """

        m = ( self._p ** ( self._q + 1 ) - 1) / ( self._p - 1 )
        f = self.factor( m )
        return f

def test( ):
    todo = range( 50, 60 ) 
    todo = range( 2, 200 ) 
    for q in todo:
        p = 2
        factors = SigmaFactor( p, q )
        print "( %d, %d ) %s" % ( p, q, factors )

def main( ):

    result = None
    globals.quiet = 1

    for arg in sys.argv[1:]:

        if arg[ 0 ] == 'l':
            global verbose
            verbose = 0
            globals.PROG.load( ) 
            continue

        if arg[ 0 ].isdigit( ):
            ( p, q, r ) = FromString( arg )
            print "%s^%s  %s" % ( p, q, r )
            if result == None:
                result = r
            else:
                result = result * r
            continue

        if arg[ 0 ] == 's':
            verbose = 0
            globals.PROG.doSave( True ) 
            continue

        if arg[ 0 ] == 't':
            test( )

    if result:
        print result

if __name__ == '__main__':
    main( )

