###!/usr/bin/python

import globals

import re, time
import pydb

import Cache
import Factors
import Number
import TimeStamp

class SigmaFactorCache( Cache.Cache ):

    def __init__( self, saver = None ):
        Cache.Cache.__init__( self, saver )

    def type( self ):

        return "SigmaFactorCache"

    def __str__( self ):

        quiet = globals.quiet
        globals.quiet = False

        answer = ""
        keys = self.sortedkeys( )
        for key in keys:
            p = key.prime( )
            q = key.power( )
            ( d, f ) = self._dict[ key ]
            if globals.quiet > 0:
                answer += "%s  %s  %s\n" % ( p, q, f )
            else:
                answer += "%s    %s  %s  %s\n" % ( d, p, q, f )

        globals.quiet = quiet

        if answer == "":
            answer = "<SigmaFactorCache.SigmaFactorCache instance>"

        return answer

    def addString( self, line ):
        tokens = re.split( "  +", line )
        date = TimeStamp.TimeStamp( tokens[ 0 ] )

        prime = tokens[ 1 ]
        power = tokens[ 2 ]
        num = Number.FromStringPair( prime, power )

        tokens = re.split( "[()]", line )
        factors = Factors.FromString( tokens[ 1 ] )

        self.addValue( num, factors, date, False )

        return

    def addValue( self, key, factors, date = None, saver = True ):
        """
        key is a Number
        factors is a Factor list
        """

        if self.verify( key, factors ):
            if not self._dict.has_key( key ):
                if date == None:
                    date = TimeStamp.TimeStamp( )
                self._dict[ key ] = ( date, factors.copy( ) )
                if saver:
                    globals.Saver.newSgmftr( )

        return

    def at( self, key ):
        try:
            ( date, value ) = self._dict[ key ]
            return value.copy( )
        except:
            return None

    def verify( self, value, factors ):

        prime = value.prime( )
        power = value.power( )

        self_mag = self.sgmftr( prime, power )
        fact_mag = factors.magnitude( )

        if self_mag == fact_mag:
            return True
        else:
            print "SigmaFactorCache error:"
            print "%s^%s %s" % ( num.prime( ), num. power( ), factors )
            pydb.debugger( )
            return False

    def sgmftr( self, p, q ):
        # Compute sgmftr(p,q)
        nn = ( p ** ( q + 1 ) - 1 ) / ( p - 1 )
        return nn
