#!/usr/bin/python

import globals

import Number

class StartTripleDict( dict ):
    """A dictionary mapping a prime to a StartTriple instantiated with
    the prime, a power, and min and max power values
    """

    def __init__( self, primes ):
        """Initialize using an array of primes
        """
        
        self._dict = { }
        self._primes = [ ]

        self.addTriple( Number.Number( 2 ) )

        for prime in primes:
            self.addTriple( Number.Number( prime ) )

        return

    def type( self ):

        return "StartTripleDict"

    def __str__( self ):

        answer = ""
        for p in self._primes:
            if p != 2:
                answer += "%s " % self._dict[ p ]

        return answer

    def addTriple( self, factor, min = 0, max = 0 ):

        p = factor.prime( )
        q = factor.power( )
        s = StartTriple( p, q, min, max )
        self._dict[ p ] = s

        self._primes.append( p )
        self._primes.sort( )

        return

    def primes( self ):
        """
        return sorted list of primes (keys)
        """

        return self._primes

    def xxx_sortedkeys_xxxdef( self, key = None ):

        if not key:
            values = self._dict.values( )
        else:
            values = self._dict[ key ].values( )
        
        values.sort( )

        return values

    def sortedvalues( self ):

        values = self._dict.values( )
        
        values.sort( )

        return values

    def item_number( self, index ):

        keys  = self._dict.keys( )
        key   = keys[ index ]
        value = self[ key ]

        return value
    
    def first( self ):
        """
        Answer the receiver's first item.
        """

        p = self._primes[ 0 ]
        v = self._dict[ p ]

        return v

    def count( self ):

        count = 1
        for key in self._primes:
            triple = self._dict[ key ]
            if triple.prime( ) != 2:
                count = count * ( triple.max( ) - triple.min( ) + 1 )

        return count

    def display( self ):

        print "%s %d" % ( self, self.count( ) )

        return

    def updateStartTriples( self, seed ):
        # search result cache for the same power of two
        # and set up/lower limits for triples' primes

        self.updateFromSeed( seed )

        return

    def updateFromSeed( self, seed ):
        # search result cache for the same power of two
        # and set up/lower limits for triples' primes

        globals.ResultCache.updateStartTriples( self._dict, seed )

        return

    def updateFromFactors( self, factors ):
        # search result cache for the same power of two
        # and set up/lower limits for triples' primes

        """
        find matching power of 2 in result cache
        find min & max powers of 3, 5, and 7
        """

        triples = self._dict
        keys = triples.keys( )

        for p in keys:
            if factors.has_key( p ):
                factor = factors[ p ]
                triple = triples[ p ]
                triple.adjMinMax( factor )

        return

    def widen( self, min, max ):

        for triple in self._dict.values( ):
            triple.widen( min, max )

        return

class StartTriple:
    """A prime, a power, and min and max values
    """

    def __init__( self, prime, power, lo, hi ):

        self._prime = prime
        self._power = power
        if not lo or not hi:
            self._min = 0
            self._max = 0
        else:
            if lo >= hi:
                lo = hi - 1
            self._min = max( lo, 0 )
            self._max = max( hi, 0 )

        return

    def type( self ):

        return "StartTriple"

    def initMinMax( self, min, max ):
        "Initialize instance variables min and max."
        "Used by AlbTripleStart::newFirstPower2."

        self._power = 0
        if self._min == 0  or self._min > min:
            self._min = min
        if self._max == 0  or self._max < max:
            self._max = max

        return

    def adjMinMax_xxxdef( self, factor ):

        power = factor.power( )
        if not self._min or power < self._min:
            self._min = power
        if not self._max or power > self._max:
            self._max = power

    def decrementPower( self ):
        "Decrement instance variable power."

        self._power -= 1

        return self._power

    def finalizeMinMax( self ):
        "Finalize instance variables min and max."

        self._min = max( 0, self._min - 2 )
        self._max = self._max + 2

        return

    def incrementMax( self, d = 1 ):
        "Increment instance variable max."

        self._max = self._max + d

        return self._max
 
    def incrementPower( self, d = 1 ):
        "Increment instance variable power."

        if self._power < 0:
            self._power = self._min
        elif self._power == 0 and self._min != 0:
            self._power = self._min
        else:
            self._power += d

        return self._power
  
    def widen( self, min, max ):

        '''
        decrease min and increase max
        '''

        if self._min > min:
            self._min -= min
        else:
            self._min = 1
        self._max += max

        return self

    def max( self, anObject = None ):
        "Get/Set the instance variable max."

#       if anObject or anObject == 0:
#       if anObject != None and ( self._max == 0 or anObject > self._max ):
        if anObject:
            if ( self._max == 0 or anObject > self._max ):
                self._max = anObject

        return self._max
   
    def min( self, anObject = None ):
        "Get/Set the instance variable min."

#       if anObject or anObject == 0:
#       if anObject != None and ( self._min == 0 or anObject < self._min ):
        if anObject:
            if ( self._min == 0 or anObject < self._min ):
                self._min = anObject

        return self._min
   
    def prime( self, anObject = None ):
        "Get/Set the instance variable prime."

        if anObject:
            self._prime = anObject

        return self._prime
   
    def power( self, anObject = None ):
        "Get/Set the instance variable power."

        if anObject or anObject == 0:
            self._power = anObject

        return self._power
   
    def primeRatio( self ):

        return int( self._power * 3 / 5 )
   
    def __cmp__( self, other ):
        """compare two StartTriples
        if their primes are different,
        return the difference of the primes
        else return the difference of the powers
        """
        if self._prime != other._prime:
            return self._prime - other._prime
        else:
            return self._power - other._power

    def __repr__( self ):

        return self.asString( )

    def __str__( self ):

        return self.asString( )

    def asString( self ):

        answer = "( %s, %s (%s:%s))" % ( self._prime, self._power, self._min, self._max )

        return answer

    def updateMinMax( self, anInteger):
        "Update the receiver's exponent range."

        if self._min == 0 or self._min > anInteger:
            self._min = anInteger
        if self._max == 0 or self._max < anInteger:
            self._max = anInteger

        return
    
    def setExponentRange( self, anInteger):
        "Update the receiver's span (minimum exponent range)."

        if anInteger == 0:
            return

        diff = self._max - self._min
        if diff < anInteger:
            delta = ( anInteger - diff ) / 2
            self._min -= delta
            self._max += delta
            # if less than 0, use 0..anInteger
            if self._min < 0:
                self._min = 0
                self._max = anInteger

        return
    
if __name__ == '__main__':

    if len( sys.argv ) == 1:

        alb = ALB( )
        alb.load( filename )
        
        primes = [ 3, 5, 7, 11 ]
        
    else:
        for arg in sys.argv[1:]:

            if arg[ 0 ] == 't':
                TestStarter( )

            if arg[ 0 ] == 'l':
                TestLoader( "RSLTS.txt" )
