#!/usr/bin/python

# Sat Jun  9 21:55:33 EDT 2007
# ./ALB.py RSLTS.ALB l s1.1000
#    counts:  NEW - 0,  OLD - 335,  DUP - 99, TOT - 434
# time: 14.35

import globals
import pydb
import sys, re
from mpz import *

import AlbDict
import Factors
import Number
import NumDenMul
import StartTriple

class TripleStart:

    def __init__( self, mode = None, two = None, seed = None ):

        self._ranges = AlbDict.AlbDict( )
        
        self.initRanges( )
        self._position = 0
        self._maxCol = 4
        self._maxRow = 250
        self._tst_cnt = 0

#       SearchBasedOnLowCount
        if mode == "lc":
            triples = self.findResultMinMax( seed )
            triples.widen( 0, 3 )

            self.set_triples( two, triples )
            self.setRange( two.prime( ), two.power( ) )

#       SearchBasedOnFactorReduction
        if mode == "fr":
            triples = self.findResultAndFactorReduce( seed )
            triples.widen( 0, 2 )

            self.set_triples( two, triples )
            self.setRange( two.prime( ), two.power( ) )

        return

    def type( self ):

        return "TripleStart"

    def initialize( self ):

        self._first = True

        return self

    def __str__( self ):

        #tri : max(row,col) = (250,3),  tst_cnt = 1000 
        #tri : ( (2,9 (1:100)) (3,10 (1:12)) (5,0 (1:9)) )
        #counts:  NEW - 0,  OLD - 19,  DUP - 19, TOT - 275

        answer = "tri : max(row,col) = (%d,%d), tst_cnt = %d\n" % ( self._maxRow, self._maxCol, self._tst_cnt )
        answer += "tri : %s" % self.ranges_as_string( )

        return answer

    def ranges_as_string( self ):

        answer = "("
        for range in self._ranges.sortedvalues( ):
            answer += "%s " % range
        answer += ")"

        return answer

    def display( self ):

        print "%s %d" % ( self.rangeString( ), self.rangeCount( ) )

    def rangeCount( self ):

        count = 1
        total = 0

        for range in self._ranges.sortedvalues( ) :
            if range.prime( ) != 2:
                count = count * ( range.max( ) - range.min( ) + 1 )
                total = total + count

        return total

    def rangeString( self, pfx = None ):

        result = ""

        for range in self._ranges.sortedvalues( ) :
            if pfx: result +=  pfx
            if range.prime( ) != 2:
                result += "%s " % range

        return result

    def load( self, line ):
        # tri : max(row,col) = (250,3), tst_cnt = 0
        # tri : (( 2, 1 (1:20)) ( 3, 0 (0:0)) ( 5, 0 (0:0)) )
        
        parts = re.split( "[ (),=:]+", line )
        if not parts[3][0].isdigit( ):
            self.loadRowColCnt( parts )
        else:
            self.loadRanges( parts )

    def loadRowColCnt( self, parts ):

        self._maxRow = int( parts[ 4 ] )
        self._maxCol = int( parts[ 5 ] )
        self._tst_cnt = int( parts[ 7 ] )

        return    

    def loadRanges( self, parts ):

        self._ranges = AlbDict.AlbDict( ) # dict

        for index in range( 1, len( parts ) - 1, 4 ):
            prime = mpz( parts[ index + 0 ] )
            power = mpz( parts[ index + 1 ] )
            min   = mpz( parts[ index + 2 ] )
            max   = mpz( parts[ index + 3 ] )
            self.setRange( prime, power, min, max )

        return    

    def asFactorList( self ):
        """
        Answer a factor list.
        """

        answer = Factors.Factors( None )
        for value in self._ranges.sortedvalues( ):
            p = value.prime( )
            q = value.power( )
            if p == 2:
                answer.add( Number.Number( p, max( 1, q ) ) )
            elif q > 0:
                answer.add( Number.Number( p, q ) )

        return answer

    def at( self, anInteger ):
        """
        Answer an element of the list.
        """

        return self._ranges[ anInteger ]

    def prime_power( self, prime, power ):
        starter = StartTriple.StartTriple( prime, 0, 0, power )
        self._ranges[ prime ] = starter
        return

    def cycleCount( self ):
        "Display the receiver."

        count = 1

        for triple in self._ranges.sortedvalues( ):
            count = count * ( triple.max ( ) - triple.min( ) + 1 )

        return count 
        
    def initRanges( self ):
        "Initialize with a default list of primes."

        primes = [ 2, 3, 5, 7, 11, 13, 17, 19 ]
        primes = [ 2, 3 ]

        for prime in primes:
            self.setRange( prime, 0 )

    def firstExponent( self ):
        """
        Answer the first exponent
        """

        range = self._ranges.first( )

        return range.power( )

    def initializePrimes( self ):
        """
        Initialize the receiver.
        """

        self._ranges = self._ranges.copy( 1, self._maxcol )
        self.second( ).decrementpower( )

        return

    def newFirstPower( self ):
        # from AlbLimitedTripleStart
        """
        Update the exponent of the first rangeList item.
        """

        first = True
        for triple in self._ranges.sortedvalues( ):
            if first:
                first = False
                power = triple.incrementPower( )
            else:
                power = power * 3 / 5 
                triple.initMinMax( 0, power )

        results = None
        if results:
            for result in results:
                for factors in result:
                    index = self.rangesIndexOfPrimeValue( selfe._ranges, factor.prime( ) )
                    if index != None and factor.prime( ) != 2:
                        range.setMinMax( factor.power( ) )

        return

    def rangesIndexOfPrimeValue( self, ranges, prime ):
        for range in ranges:
            if range.prime( ) == prime:
                return range
        return None

    def asString( self ):
        return self.__str__( )

    def count( self ):
        return self._tst_cnt

    def next( self ):
        """
        Answer a factor list.
        """

        if self.firstExponent( ) >= self._maxRow:
            return None

        self._tst_cnt += 1
        self.updateExponents( )
        value = self.value( )

        return ( self._tst_cnt, value )

    def primes( self, integers ):
        """
        Save the supplied array of integers as an array of AlbStartTriples.
        """

        self._ranges = AlbDict.AlbDict( ) # dict

        for index in range( self._maxCol ):
            number = integers[ index ]
            self._ranges[ number ] = StartTriple.StartTriple( number, 0, 0, 0 )
        first = self._ranges[ 0 ]
        first.max( 30 / first.prime( ) )

        return

    def __getitem__( self, prime ):

        try:
            ans = self._ranges[ prime ]
        except:
            if prime < 20:
                ans = self.setRange( prime, 0 )
                print ans

        return ans

    def __setitem__( self, prime, factor ):

        self._ranges[ prime ] = factor

        return factor

    def mergeResults( self, power, results_set ):

        for result in results_set:
            factors = result.factors( )

            self.mergeFactors( factors )

        return self

    def mergeFactors( self, factors ):

        '''
        merge factors that match the default list of primes
        '''
        
        factors = factors.factors( )
        for factor in factors:
            p = factor.prime( )
            q = factor.power( )
            if self._ranges.has_key( p ):
                triple = self[ p ]
                triple.min( q )
                triple.max( q )
            elif p < 20:
                self.setRange( p, 0, q, q )

        return self

    def setRange( self, prime, power, min = None, max = None ):

        if not min:
            min = power
        if not max:
            max = power

        starter = StartTriple.StartTriple( prime, power, min, max )
        self._ranges[ prime ] = starter

        return starter

    def ranges( self ):
        """
        Answer the rangeList.
        """

        return self._ranges

    def set_triples( self, two, triples ):

        range = self._ranges[ 2 ]
        range.power( two.power( ) )

        dict = triples._dict
        for triple in dict.values( ):
            self._ranges[ triple.prime( ) ] = triple

        return self

    def size( self ):
        """
        Answer the number of elements in the receiver's range list.
        """

        return len( self._ranges )

    def clearExponents_XXX( self, start ):
        for index in range( start, self.size( )  ):
            triple = self._ranges[ index ]
            triple.power( 0 )

    def updateExponents( self ):
        """
        Update exponents.
        """
        
        if self._first:
            self._first = False
            return

        for triple in self._ranges.sortedvalues( ):
            if triple.prime( ) == 2:
                continue
            if triple.power( ) < triple.max( ):
                triple.incrementPower( )
                return
            else:
                triple.power( 0 )
#                triple.power( triple.min( ) )

        self.newFirstPower( )

        return

    def adjMinMax_xxxdef( self, seed, factors ):

        """factors is ndm.num() for an invocation of ALB::cycle
        for a new power of 2.
        """

        starters = self.findResultMinMax( seed )
        starters.display( )
        for factor in seed.values( ):
            starters.addTriple( factor )
        self._ranges = starters

        if False:    # 0318 #
            for range in self._ranges.sortedvalues( ):
                if range.prime( ) != 2:
                    range.min( 0 )
                    range.max( 0 )

            keys = self._ranges.primes( )

            range_max = max( keys )
            self.adjSpan( starters, factors, 8, range_max )

        if False:    # 0317
            pow_two = two.power( )
            power = min( pow_two + 5, 10 )
            for range in self._ranges.sortedvalues( ):
                if range.prime( ) == 2:
                    continue
                if range.max( ) != 0:
                    power = range.max( )
                else:
                    power = power * 7 / 10
                    range.max( power )

        return

    def adjSpan( self, starters, factors, span, range_max ):

        for factor in factors.sortedvalues( ):
            p = factor.prime( )
            q = factor.power( )
            if p == 2:
                continue
            if p > range_max:
                continue
            if not self._ranges.has_key( p ):
                continue
            r = self._ranges[ p ]
            if q < 1:
                val = 0
            else:
                val = int( q ) - 1

            if starters.has_key( p ):
                starter = starters[ p ]
                if starter.min( ) != 0:
                    r.min( max( 1, starter.min( ) - 1 ) )
                    r.max(         starter.max( ) + 1   )
                    r.setExponentRange( span )
                continue

            r.min( val )
            val = min( val + span, pow_two + 5 )
            r.max( val )
            r.setExponentRange( span )
#           span = span * 7 / 10

        return

    def findResultMinMax( self, seed ):
        """
        find matching power of 2 in result cache
        find min & max powers of 3, 5, 7, ...
        """

        triples = self.findResultSetup( seed )

        # search result cache for the same power of two
        # and set upper/lower limits for triples' primes
        triples.updateFromSeed( seed )
        triples.widen( 0, 2 )

        return triples
                    
    def findResultAndFactorReduce( self, seed ):

        triples = self.findResultSetup( seed )

        ndm = NumDenMul.new( seed )
        max = 17
        ndm = ndm.reduceLargeFactors( max )

        factors = ndm.num( )
        triples.updateFromFactors( factors )

        return triples

    def findResultSetup( self, seed ):
        """
        find matching power of 2 in result cache
        find min & max powers of 3, 5, 7, ...
        """

        if globals.printDate:
            print "%s" % TimeStamp.TimeStamp( ),

        print
        print "*** %s ***" % seed

        primes = [ 3, 5, 7, 11, 13, 17, 19 ]

        triples = StartTriple.StartTripleDict( primes )

        return triples
                    
    def updateUpperLimits_XXX( self ):
        """
        Update the upper limits of the range list.
        """

        pydb.debugger( )
        for index in range( self._position, self.size( ) - 1 ):
            cur = self.at( index )
            nxt = self.at( index + 1 )
            limit = cur.primeRatio( )
            if limit > nxt.max( ):
                nxt.max( limit )

        return

    def value( self ):
        """
        Answer a factor list.
        """

        return self.asFactorList( )

    def widen( self, min, max ):

        '''
        decrease min and increase max
        '''

        for range in self._ranges.sortedvalues( ):
            if range.prime( ) != 2:
                range.widen( min, max )

        return self

def TestStarter_1( ):
    print "TripleStart"

    starter = TripleStart( "lc", Number.Number(2,3) )

    print "ts: %s" % starter
    print ""

    for index in range( 1, 100 ):
        value = starter.next( )
        print "in: %d .%s" % ( index, value )

    return

def TestLoader( filename ):

    starter=TripleStart()

    file = open( filename, "r" )

    line = file.readline( )
    starter.load( line )

    line = file.readline( )
    starter.load( line )

    file.close( )

    print "%s" % starter

    return

def TestStarter_2( ):

    # ( 2, 0 (0:1000)) ( 3, 0 (0:0)) ( 5, 0 (0:0)) ( 7, 0 (0:0))
    starter = TripleStart( )
    for index in range( 1, 10 ):
        # ( 2, 1 ... ) ...
        starter.newFirstPower( )
        print "%d %s" % (index, starter.rangeString( ) )

    p = starter._ranges[ 2 ].prime( )
    q = starter._ranges[ 2 ].power( )

    pydb.debugger( )

#   two = Number.Number( 2, arg )
    two = Number.Number( p, q )
    ndm = NumDenMul.new( seed )
    factors = ndm.reduceLargeFactors( 10 )
    factors = Factors.Factors( p ** q - 1 )
    starter.adjMinMax( two, factors )
    print starter.rangeString( )

    for count in range( 1, 40 ):
        ( cnt, value ) = starter.next( )
        print "%2d %d %s" % ( count, cnt, value )
        
    return

if __name__ == '__main__':

# scenario:  TripleStart.py l 17 r

    if len( sys.argv ) == 1:
        TestLoader( "RSLTS.19XX.txt" )
    else:
        for arg in sys.argv[1:]:

            # TestStarter
            if arg[ 0 ] == 't':
                TestStarter_1( )

            # load RESULTS.txt
            if arg[ 0 ] == 'l':
                import ALB
                alb = ALB.ALB( )
                alb.load( )

            if arg[ 0 ].isdigit( ):
                two = Number.Number( 2, arg )

                starter = TripleStart( )
                triples = starter.findResultMinMax( two )
                
                starter.set_triples( two, triples )

                print "*** starter ***"
                starter.display( )
                
            # run
            if arg[ 0 ] == 'r':
                cnt = 1

                while cnt < 1000:
                    cnt += 1

                    next = starter.next( )
                    if next == None:
                        break

                    ( test_count, factors ) = next

#                    seed = factors
#                    ndm = NumDenMul.new( seed )

                    if len( factors ) == 1:
#                        ndm_seen = set( )
                        starter.adjMinMax( two, factors )

                    print "%3d %s" % ( cnt, factors )

