#!/usr/bin/python

import gmpy

""" 
gmpy.mpz() values created from ints work properly as dictionary keys

when a gmpy.mpz() value created from a long is
used as a dictionary key, an exception is thrown.

if the value is cast to a long, it can be used as a key.
"""

# dict problem with gmpy.mpz objects as keys
#      OK with cast to long

def test_int_keys( ):

    print "#### test_int_keys ####"
    print

    dict = { }
    
    num = 0x10000FFFFFF
    val = 73

    for idx in range( 1, 4 ):
        num = num * 16 + 0xF
        print "0x%X %d" % ( num, num )
        key = gmpy.mpz( int( num ) )
        try:
            dict[ key ] = val
            print "'set' ok"
        except:
            dict[ long( key ) ] = val
            print "'set' error"
        try:
            tmp = dict[ key ]
            print "'get' ok"
        except:
            tmp = dict[ long( key ) ]
            print "'get' error"
        print

def test_long_keys( ):

    print "#### test_long_keys ####"
    print

    dict = { }

    num = 0x10000FFFFFF
    val = 73

    for idx in range( 1, 4 ):
        num = num * 16 + 0xF
        print "0x%X %d" % ( num, num )
        key = gmpy.mpz( long( num ) )
        try:
            dict[ key ] = val
            print "'set' ok"
        except:
            dict[ long( key ) ] = val
            print "'set' error"
        try:
            tmp = dict[ key ]
            print "'get' ok"
        except:
            tmp = dict[ long( key ) ]
            print "'get' error"
        print

def test_string_keys( ):

    print "#### test_string_keys ####"
    print

    dict = { }

    num = 0x10000FFFFFF
    val = 73

    for idx in range( 1, 4 ):
        num = num * 16 + 0xF
        print "0x%X %d" % ( num, num )
        key = "%s" % num
        key = gmpy.mpz( key )
        try:
            dict[ key ] = val
            print "'set' ok"
        except:
            dict[ long( key ) ] = val
            print "'set' error"
        try:
            tmp = dict[ key ]
            print "'get' ok"
        except:
            tmp = dict[ long( key ) ]
            print "'get' error"
        print

def test_svn( ):
    h1 = hash(long('12345678901234567890'))  
    h2 = hash(gmpy.mpz('1234567901234567890'))  
    print h1, h2

#!/usr/bin/python

import gmpy

def pollard_rho_factors_gmpy( value ):
#    print value
    value = gmpy.mpz( value )
    print "%e" % ( value * 1.0 )

def test_pollard_rho_factors_gmpy( ):
    
    pollard_rho_factors_gmpy( "1315127849590173" )
    pollard_rho_factors_gmpy( "6871885176561461568863874645493002864176010" )
    pollard_rho_factors_gmpy( "5225260318375411481909498130" )
    pollard_rho_factors_gmpy( "27303345394788706562827484688393677076853482354201991293" )
    pollard_rho_factors_gmpy( "572484745741528457369486524730490" )
    pollard_rho_factors_gmpy( "107412710607765960615641571053818523598097972001398330818137211296387637292222366626202740890544309751111023685652112874804014960905" )
    
if __name__ == '__main__':

    test_int_keys( )
    test_long_keys( )
    test_string_keys( )
    test_svn( )
    test_pollard_rho_factors_gmpy( )
    
